
from enum import Enum as _Enum

__all__ = ["TransactionInfo", "TransactionCode"]


class TransactionCode(_Enum):
    CREDIT = "CR"
    DEBIT = "DR"
    CURRENT_LIABILITY = "CL"
    ACCOUNT_RECEIVABLE = "AR"
    RECEIVED_RECEIPT = "RR"
    SENT_RECEIPT = "SR"
    RECEIVED_REFUND = "RF"
    SENT_REFUND = "SF"


class TransactionInfo:
    """This class is used to encode and extract the type of transaction
       and value to/from an object store key
    """
    def __init__(self, key=None):
        """Construct, optionally from the passed key"""
        if key is not None:
            t = TransactionInfo.from_key(key)

            import copy as _copy
            self.__dict__ = _copy.copy(t.__dict__)
        else:
            from Acquire.Accounting import create_decimal as _create_decimal
            self._value = _create_decimal(0)
            self._receipted_value = _create_decimal(0)
            self._code = None
            self._datetime = None
            self._uid = None

    def __str__(self):
        if self._receipted_value is None:
            return "TransactionInfo(code==%s, value==%s)" % \
                        (self._code.value, self._value)
        else:
            return "TransactionInfo(code==%s, value==%s, receipted==%s)" % \
                        (self._code.value, self._value, self._receipted_value)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._code == other._code and \
                   self._value == other._value
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def _get_code(code):
        """Return the TransactionCode matching 'code'"""
        return TransactionCode(code)

    @staticmethod
    def encode(code, value, receipted_value=None):
        """Encode the passed code and value into a simple string that can
           be used as part of an object store key. If 'receipted_value' is
           passed, then encode the receipted value of the provisional
           transaction too
        """
        if receipted_value is None:
            return "%2s%013.6f" % (code.value, value)
        else:
            return "%2s%013.6fT%013.6f" % (code.value, value, receipted_value)

    def rescind(self):
        """Return a TransactionInfo that corresponds to rescinding this
           transaction info. This is useful if you want to update the
           ledger to remove this object (since we can't delete anything
           from the ledger)
        """
        t = TransactionInfo()
        t._uid = self._uid[-1::-1]
        t._value = self._value
        t._datetime = self._datetime

        if self._code is TransactionCode.DEBIT:
            t._code = TransactionCode.CREDIT
        elif self._code is TransactionCode.CREDIT:
            t._code = TransactionCode.DEBIT
        elif self._code is TransactionCode.CURRENT_LIABILITY:
            t._value = -(self._value)
        elif self._code is TransactionCode.ACCOUNT_RECEIVABLE:
            t._value = -(self._value)
        else:
            raise PermissionError(
                "Do not have permission to rescind a %s" % str(self))

        return t

    def value(self):
        """Return the value of the transaction. This will be the receipted
           value if this has been set"""
        if self._receipted_value is not None:
            return self._receipted_value
        else:
            return self._value

    def uid(self):
        """Return the UID of this transaction"""
        return self._uid

    def dated_uid(self):
        """Return the full dated uid of the transaction. This
           is isoformat(datetime)/uid
        """
        return "%s/%s" % (self._datetime.isoformat(), self._uid)

    def datetime(self):
        """Return the datetime of this transaction"""
        return self._datetime

    @staticmethod
    def from_key(key):
        """Extract information from the passed object store key.

           This looks for a string that is;

           isoformat_datetime/UID/transactioncode

           where transactioncode is a string that matches
           '2 letters followed by a number'

           CL000100.005000
           DR000004.234100

           etc.

           For sent and received receipts there are two values;
           the receipted value and the original estimate. These
           have the standard format if the values are the same, e.g.

           RR000100.005000

           however, they have original value T receipted value if they are
           different, e.g.

           RR000100.005000T000090.000000

           Args:
                key: Object store key

        """
        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime
        from Acquire.Accounting import create_decimal as _create_decimal

        parts = key.split("/")

        # start at the end...
        nparts = len(parts)
        for i in range(0, nparts):
            j = nparts - i - 1
            t = TransactionInfo()

            try:
                t._datetime = _string_to_datetime(parts[j-2])
            except:
                continue

            t._uid = parts[j-1]

            part = parts[j]
            try:
                code = TransactionInfo._get_code(part[0:2])

                if code == TransactionCode.SENT_RECEIPT or \
                   code == TransactionCode.RECEIVED_RECEIPT:
                    values = part[2:].split("T")
                    try:
                        value = _create_decimal(values[0])
                        receipted_value = _create_decimal(values[1])
                        t._code = code
                        t._value = value
                        t._receipted_value = receipted_value
                        return t
                    except:
                        pass

                value = _create_decimal(part[2:])

                t._code = code
                t._value = value
                t._receipted_value = None

                return t
            except:
                pass

        raise ValueError("Cannot extract transaction info from '%s'"
                         % (key))

    def to_key(self):
        """Return this transaction encoded to a key"""
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string

        return "%s/%s/%s" % (_datetime_to_string(self._datetime),
                             self._uid,
                             TransactionInfo.encode(
                                    code=self._code,
                                    value=self._value,
                                    receipted_value=self._receipted_value))

    def receipted_value(self):
        """Return the receipted value of the transaction. This may be
           different to value() when the transaction was provisional,
           and the receipted value is less than the provisional value.
           This returns None if this transaction wasn't receipted

           Returns:
                Decimal: Receipted value of Transaction
        """
        return self._receipted_value

    def original_value(self):
        """Return the original (pre-receipted) value of the transaction"""
        return self._value

    def is_credit(self):
        """Return whether or not this is a credit

           Returns:
                bool: True if this is a credit, else False
        """
        return self._code == TransactionCode.CREDIT

    def is_debit(self):
        """Return whether or not this is a debit

           Returns:
                bool: True if this is a debit, else False

        """
        return self._code == TransactionCode.DEBIT

    def is_liability(self):
        """Return whether or not this is a liability

           Returns:
                bool: True if this is a liability, else False

        """
        return self._code == TransactionCode.CURRENT_LIABILITY

    def is_accounts_receivable(self):
        """Return whether or not this is accounts receivable

           Returns:
                bool: True if this is accounts receivable, else False

        """
        return self._code == TransactionCode.ACCOUNT_RECEIVABLE

    def is_sent_receipt(self):
        """Return whether or not this is a sent receipt

           Returns:
                bool: True if this is accounts receivable, else False

        """
        return self._code == TransactionCode.SENT_RECEIPT

    def is_received_receipt(self):
        """Return whether or not this is a received receipt

           Returns:
                bool: True if this is a received receipt, else False
        """
        return self._code == TransactionCode.RECEIVED_RECEIPT

    def is_sent_refund(self):
        """Return whether or not this is a sent refund

           Returns:
                bool: True if this is a sent refund, else False
        """
        return self._code == TransactionCode.SENT_REFUND

    def is_received_refund(self):
        """Return whether or not this is a received refund

           Returns:
                bool: True if this is a received refund, else False
        """
        return self._code == TransactionCode.RECEIVED_REFUND




__all__ = ["get_filesize_and_checksum", "get_size_and_checksum"]


def get_size_and_checksum(data):
    """Calculates the size and md5 of the passed data

       Args:
            data (byte): data to calculate checksum for
        Returns:
            tuple (int,str): size of data and its md5 hash
    """
    from hashlib import md5 as _md5
    md5 = _md5()
    md5.update(data)

    return (len(data), str(md5.hexdigest()))


def get_filesize_and_checksum(filename):
    """Opens the file with the passed filename and calculates
        its size and md5 hash

       Args:
            filename (str): filename to calculate size and checksum for
        Returns:
            tuple (int,str): size of data and its md5 hash

    """

    from hashlib import md5 as _md5
    md5 = _md5()
    size = 0

    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5.update(chunk)
            size += len(chunk)

    return (size, str(md5.hexdigest()))




from Acquire.Access import Request as _Request

__all__ = ["RunRequest"]


class RunRequest(_Request):
    """This class holds a request to run a particular calculation
       on a RunService. The result of this request will be a
       Location in which the output from this request can
       be read.
    """
    def __init__(self, image=None, input=None, resources=None):
        """Construct the request specifying the container image 'image'
           that contains the software used for the calculation,
           and the location of the input files 'input' that will be
           downloaded and run using this container.

           You can also optionally supply the compute resources
           (resources) that will be needed to run this job
        """
        super().__init__()

        self._uid = None
        self._image = str(image)
        self._resources = resources

        self._input = input

        if self._input is not None:
            from Acquire.Client import Location as _Location
            if not isinstance(self._input, _Location):
                raise TypeError("The input location must be type Location")

            from Acquire.ObjectStore import create_uid as _create_uid
            self._uid = _create_uid()

    def is_null(self):
        """Return whether or not this is a null request"""
        return self._uid is None

    def __str__(self):
        if self.is_null():
            return "RunRequest::null"
        else:
            return "RunRequest(uid=%s, image=%s, input=%s)" % (
                    self._uid, self._image, self._input)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def uid(self):
        """Return the UID of this request"""
        return self._uid

    def fingerprint(self):
        """Return a unique fingerprint for this request that can be
           used for signing and verifying authorisations
        """
        if self.is_null():
            return None

        return "%s|%s|%s" % (self.uid(), self.image(),
                             self.input().to_string())

    def image(self):
        """Return the full path to the container image used for the
           simulation
        """
        if self.is_null():
            return None
        else:
            return self._image

    def input(self):
        """Return the location of the input files used for this
           calculation
        """
        if self.is_null():
            return None
        else:
            return self._input

    def resources(self):
        """Return the resources requested to run this job"""
        if self.is_null():
            return None
        else:
            return self._resources

    def to_data(self):
        """Return this request as a json-serialisable dictionary"""
        if self.is_null():
            return {}

        data = super().to_data()
        data["uid"] = self._uid
        data["input"] = self._input.to_data()

        if self._image is not None:
            data["image"] = self._image

        if self._resources is not None:
            data["resources"] = str(self._resources)

        return data

    @staticmethod
    def from_data(data):
        """Creates a RunRequest object from the JSON data in data"""
        if data and len(data) > 0:
            from Acquire.Client import Location as _Location
            r = RunRequest()
            r._from_data(data)

            r._uid = data["uid"]
            r._input = _Location.from_data(data["input"])

            if "image" in data:
                r._image = str(data["image"])

            if "resources" in data:
                r._resources = str(data["resources"])

            return r

        return None




__all__ = ["Location"]


class Location:
    """This class holds a globally-resolvable location for a
       file or Drive.
    """
    def __init__(self, drive_guid=None, filename=None, version=None):
        """Construct an Location. This uses the GUID of the
           drive to identify the drive, and then (optionally) the
           name of the file in the drive, and then the specific
           version.

           The drive GUID has the format {service_uid}/{drive_uid}

           If the filename is not supplied, then this locates
           the drive itself. If a version is not supplied, then this
           locates the latest version of the file
        """
        self._drive_guid = drive_guid
        self._encoded_filename = None
        self._filename = None
        self._version = None

        if self._drive_guid is not None:
            self._filename = filename
            self._version = version

            if self._filename is not None:
                from Acquire.ObjectStore import string_to_encoded \
                    as _string_to_encoded
                self._encoded_filename = _string_to_encoded(self._filename)
            else:
                self._encoded_filename = None

    def is_null(self):
        return self._drive_guid is None

    def __str__(self):
        if self.is_null():
            return "Location::null"
        elif self.is_drive():
            return "acquire_drive://%s" % self._drive_guid
        elif self.specifies_version():
            return "acquire_file://%s/%s/%s" % (self._drive_guid,
                                                self._encoded_filename,
                                                self._version)
        else:
            return "acquire_file://%s/%s" % (self._drive_guid,
                                             self._encoded_filename)

    def fingerprint(self):
        """Return a fingerprint that can be used to show that the
           user has authorised something to do with this location
        """
        return str(self)

    def to_string(self):
        """Return a safe string that completely describes this
           location
        """
        return str(self)

    @staticmethod
    def from_string(s):
        """Return an Location constructed from the passed string"""
        if s.startswith("acquire_file://"):
            parts = s.split("/")
            from Acquire.ObjectStore import encoded_to_string \
                as _encoded_to_string

            try:
                drive_guid = parts[-2]
                filename = _encoded_to_string(parts[-1])
                version = None
            except:
                drive_guid = parts[-4]
                filename = _encoded_to_string(parts[-3])
                version = "/".join([parts[-2], parts[-1]])

            return Location(drive_guid=drive_guid, filename=filename,
                            version=version)

        elif s.startswith("acquire_drive://"):
            parts = s.split("/")
            return Location(drive_guid=parts[-1])
        else:
            return Location()

    def is_file(self):
        """Return whether or not this is a location for a file"""
        if self.is_null():
            return False
        else:
            return self._encoded_filename is not None

    def is_drive(self):
        """Return whether or not this is a location for a drive"""
        if self.is_null():
            return False
        else:
            return self._encoded_filename is None

    def specifies_version(self):
        """Return whether or not this specifies the version of the file"""
        if self.is_null():
            return False
        else:
            return self._version is not None

    def drive_guid(self):
        """Return the GUID of the drive"""
        if self.is_null():
            return None
        else:
            return self._drive_guid

    def drive_uid(self):
        """Return the UID of the drive"""
        if self.is_null():
            return None
        else:
            return self._drive_guid.split("@")[0]

    def filename(self):
        """Return the name of the file (if this is a file)"""
        return self._filename

    def encoded_filename(self):
        """Return the URL-safe encoded filename (if this a file)"""
        return self._encoded_filename

    def version(self):
        """Return the version of the file (if this is a file and
           the version has been specified)
        """
        return self._version

    def service_uid(self):
        """Return the UID of the service that holds the File/Drive
           behind this location
        """
        if self.is_null():
            return None

        return self._drive_guid.split("@")[-1]

    def service(self):
        """Return the service that holds the File/Drive behind this
           location
        """
        if self.is_null():
            return None

        from Acquire.Service import get_trusted_service \
            as _get_trusted_service

        return _get_trusted_service(service_uid=self.service_uid())

    def service_url(self):
        """Return the URL of the service that holds the File/Drive
           behind this location
        """
        if self.is_null():
            return None

        return self.service().canonical_url()

    @staticmethod
    def from_data(data):
        """Return an Location constructed from the json-deserialised
           dictionary
        """
        if data is None or len(data) == 0:
            return Location()

        iden = Location()
        iden._drive_guid = data["drive_guid"]

        try:
            iden._encoded_filename = data["encoded_filename"]
        except:
            return iden

        from Acquire.ObjectStore import encoded_to_string \
            as _encoded_to_string
        iden._filename = _encoded_to_string(iden._encoded_filename)

        try:
            iden._version = data["version"]
        except:
            pass

        return iden

    def to_data(self):
        """Return a json-serialisable dictionary of this object"""
        if self.is_null():
            return None

        data = {}
        data["drive_guid"] = self.drive_guid()

        if self.is_file():
            data["encoded_filename"] = self.encoded_filename()

        if self.specifies_version():
            data["version"] = self.version()

        return data




import io as _io
import datetime as _datetime
import uuid as _uuid
import json as _json
import os as _os
import copy as _copy
import uuid as _uuid

__all__ = ["OCI_ObjectStore"]


def _sanitise_bucket_name(bucket_name):
    """This function sanitises the passed bucket name. It will always
        return a valid bucket name. If "None" is passed, then a new,
        unique bucket name will be generated

        Args:
            bucket_name (str): Bucket name to clean
        Returns:
            str: Cleaned bucket name

        """

    if bucket_name is None:
        return str(_uuid.uuid4())

    return "_".join(bucket_name.split())


def _clean_key(key):
    """This function cleans and returns a key so that it is suitable
       for use both as a key and a directory/file path
       e.g. it removes double-slashes

       Args:
            key (str): Key to clean
       Returns:
            str: Cleaned key

    """
    key = _os.path.normpath(key)

    if len(key) > 1024:
        from Acquire.ObjectStore import ObjectStoreError
        raise ObjectStoreError(
            "The object store does not support keys with longer than "
            "1024 characters (%s) - %s" % (len(key), key))

        # if this becomes a problem then we will implement a 'tinyurl'
        # to shorten keys and use this function to lookup long keys

    return key


def _get_object_url_for_region(region, uri):
    """Internal function used to get the full URL to the passed PAR URI
       for the specified region. This has the format;

       https://objectstorage.{region}.oraclecloud.com/{uri}

       Args:
            region (str): Region for cloud service
            uri (str): URI for cloud service
       Returns:
            str: Full URL for use with cloud service
    """
    server = "https://objectstorage.%s.oraclecloud.com" % region

    while uri.startswith("/"):
        uri = uri[1:]

    return "%s/%s" % (server, uri)


def _get_driver_details_from_par(par):
    """Internal function used to get the OCI driver details from the
       passed OSPar (pre-authenticated request)

       Args:
            par (OSPar): PAR holding details
        Args:
            dict: Dictionary holding OCI driver details
    """
    from Acquire.ObjectStore import datetime_to_string \
        as _datetime_to_string

    import copy as _copy
    details = _copy.copy(par._driver_details)

    if details is None:
        return {}
    else:
        # fix any non-string/number objects
        details["created_datetime"] = _datetime_to_string(
                                        details["created_datetime"])

    return details


def _get_driver_details_from_data(data):
    """Internal function used to get the OCI driver details from the
       passed data

       Args:
            data (dict): Dict holding OCI driver details
       Returns:
            dict: Dict holding OCI driver details
    """
    from Acquire.ObjectStore import string_to_datetime \
        as _string_to_datetime

    import copy as _copy
    details = _copy.copy(data)

    if "created_datetime" in details:
        details["created_datetime"] = _string_to_datetime(
                                            details["created_datetime"])

    return details


class OCI_ObjectStore:
    """This is the backend that abstracts using the Oracle Cloud
       Infrastructure object store
    """

    @staticmethod
    def create_bucket(bucket, bucket_name, compartment=None):
        """Create and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. This will raise an
           ObjectStoreError if this bucket already exists

           Args:
            bucket (dict): Bucket to hold data
            bucket_name (str): Name of bucket to create
            compartment (str): Compartment in which to create bucket

           Returns:
                dict: New bucket
        """
        new_bucket = _copy.copy(bucket)

        new_bucket["bucket_name"] = str(bucket_name)

        if compartment is not None:
            new_bucket["compartment_id"] = str(compartment)

        try:
            from oci.object_storage.models import CreateBucketDetails as \
                _CreateBucketDetails
        except:
            raise ImportError(
                "Cannot import OCI. Please install OCI, e.g. via "
                "'pip install oci' so that you can connect to the "
                "Oracle Cloud Infrastructure")

        try:
            request = _CreateBucketDetails()
            request.compartment_id = new_bucket["compartment_id"]
            client = new_bucket["client"]
            request.name = _sanitise_bucket_name(bucket_name)

            new_bucket["bucket"] = client.create_bucket(
                                        client.get_namespace().data,
                                        request).data
        except Exception as e:
            # couldn't create the bucket - likely because it already
            # exists - try to connect to the existing bucket
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to create the bucket '%s', likely because it "
                "already exists: %s" % (bucket_name, str(e)))

        return new_bucket

    @staticmethod
    def get_bucket(bucket, bucket_name, compartment=None,
                   create_if_needed=True):
        """Find and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. If 'create_if_needed' is True
           then the bucket will be created if it doesn't exist. Otherwise,
           if the bucket does not exist then an exception will be raised.

           Args:
                bucket (dict): Bucket to hold data
                bucket_name (str): Name of bucket to create
                compartment (str, default=None): Compartment in which to
                create bucket
                create_if_needed (bool, default=None): If True, create bucket,
                else do
                not

           Returns:
                dict: New bucket

        """
        new_bucket = _copy.copy(bucket)

        new_bucket["bucket_name"] = _sanitise_bucket_name(bucket_name)

        if compartment is not None:
            new_bucket["compartment_id"] = str(compartment)

        try:
            from oci.object_storage.models import CreateBucketDetails as \
                _CreateBucketDetails
        except:
            raise ImportError(
                "Cannot import OCI. Please install OCI, e.g. via "
                "'pip install oci' so that you can connect to the "
                "Oracle Cloud Infrastructure")

        # try to get the existing bucket
        client = new_bucket["client"]
        namespace = client.get_namespace().data
        sanitised_name = _sanitise_bucket_name(bucket_name)

        try:
            existing_bucket = client.get_bucket(
                                namespace, sanitised_name).data
        except:
            existing_bucket = None

        if existing_bucket:
            new_bucket["bucket"] = existing_bucket
            return new_bucket

        if create_if_needed:
            try:
                request = _CreateBucketDetails()
                request.compartment_id = new_bucket["compartment_id"]
                request.name = sanitised_name

                client.create_bucket(namespace, request)
            except:
                pass

            try:
                existing_bucket = client.get_bucket(
                                    namespace, sanitised_name).data
            except:
                existing_bucket = None

        if existing_bucket is None:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "There is not bucket called '%s'. Please check the "
                "compartment and access permissions." % bucket_name)

        new_bucket["bucket"] = existing_bucket

        return new_bucket

    @staticmethod
    def get_bucket_name(bucket):
        """Return the name of the passed bucket

           Args:
                bucket (dict): Bucket holding data
           Returns:
                str: Name of bucket
        """
        return bucket["bucket_name"]

    @staticmethod
    def is_bucket_empty(bucket):
        """Return whether or not the passed bucket is empty

           Args:
                bucket (dict): Bucket holding data
           Returns:
                bool: True if bucket empty, else False

        """
        objects = bucket["client"].list_objects(bucket["namespace"],
                                                bucket["bucket_name"],
                                                limit=1).data

        for _obj in objects.objects:
            return False

        return True

    @staticmethod
    def delete_bucket(bucket, force=False):
        """Delete the passed bucket. This should be used with caution.
           Normally you can only delete a bucket if it is empty. If
           'force' is True then it will remove all objects/pars from
           the bucket first, and then delete the bucket. This
           can cause a LOSS OF DATA!

           Args:
                bucket (dict): Bucket to delete
                force (bool, default=False): If True, delete even
                if bucket is not empty. If False and bucket not empty
                raise PermissionError
           Returns:
                None
        """
        is_empty = OCI_ObjectStore.is_bucket_empty(bucket=bucket)

        if not is_empty:
            if force:
                OCI_ObjectStore.delete_all_objects(bucket=bucket)
            else:
                raise PermissionError(
                    "You cannot delete the bucket %s as it is not empty" %
                    OCI_ObjectStore.get_bucket_name(bucket=bucket))

        # the bucket is empty - delete it
        client = bucket["client"]
        namespace = client.get_namespace().data
        bucket_name = bucket["bucket_name"]

        try:
            response = client.delete_bucket(namespace, bucket_name)
        except Exception as e:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to delete bucket '%s'. Please check the "
                "compartment and access permissions: Error %s" %
                (bucket_name, str(e)))

        if response.status not in [200, 204]:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to delete a bucket '%s' : Status %s, Error %s" %
                (bucket_name, response.status, str(response.data)))

    @staticmethod
    def create_par(bucket, encrypt_key, key=None, readable=True,
                   writeable=False, duration=3600, cleanup_function=None):
        """Create a pre-authenticated request for the passed bucket and
           key (if key is None then the request is for the entire bucket).
           This will return a OSPar object that will contain a URL that can
           be used to access the object/bucket. If writeable is true, then
           the URL will also allow the object/bucket to be written to.
           PARs are time-limited. Set the lifetime in seconds by passing
           in 'duration' (by default this is one hour)

           Args:
                bucket (dict): Bucket to create OSPar for
                encrypt_key (PublicKey): Public key to
                encrypt PAR
                key (str, default=None): Key
                readable (bool, default=True): If bucket is readable
                writeable (bool, default=False): If bucket is writeable
                duration (int, default=3600): Duration OSPar should be
                valid for in seconds
                cleanup_function (function, default=None): Cleanup
                function to be passed to PARRegistry

           Returns:
                OSPar: Pre-authenticated request for the bucket
        """
        from Acquire.Crypto import PublicKey as _PublicKey

        if not isinstance(encrypt_key, _PublicKey):
            from Acquire.Client import PARError
            raise PARError(
                "You must supply a valid PublicKey to encrypt the "
                "returned OSPar")

        # get the UTC datetime when this OSPar should expire
        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now
        expires_datetime = _get_datetime_now() + \
            _datetime.timedelta(seconds=duration)

        is_bucket = (key is None)

        # Limitation of OCI - cannot have a bucket OSPar with
        # read permissions!
        if is_bucket and readable:
            from Acquire.Client import PARError
            raise PARError(
                "You cannot create a Bucket OSPar that has read permissions "
                "due to a limitation in the underlying platform")

        try:
            from oci.object_storage.models import \
                CreatePreauthenticatedRequestDetails as \
                _CreatePreauthenticatedRequestDetails
        except:
            raise ImportError(
                "Cannot import OCI. Please install OCI, e.g. via "
                "'pip install oci' so that you can connect to the "
                "Oracle Cloud Infrastructure")

        oci_par = None

        request = _CreatePreauthenticatedRequestDetails()

        if is_bucket:
            request.access_type = "AnyObjectWrite"
        elif readable and writeable:
            request.access_type = "ObjectReadWrite"
        elif readable:
            request.access_type = "ObjectRead"
        elif writeable:
            request.access_type = "ObjectWrite"
        else:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unsupported permissions model for OSPar!")

        request.name = str(_uuid.uuid4())

        if not is_bucket:
            request.object_name = _clean_key(key)

        request.time_expires = expires_datetime

        client = bucket["client"]

        try:
            response = client.create_preauthenticated_request(
                                        client.get_namespace().data,
                                        bucket["bucket_name"],
                                        request)

        except Exception as e:
            # couldn't create the preauthenticated request
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to create the OSPar '%s': %s" %
                (str(request), str(e)))

        if response.status != 200:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to create the OSPar '%s': Status %s, Error %s" %
                (str(request), response.status, str(response.data)))

        oci_par = response.data

        if oci_par is None:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to create the preauthenticated request!")

        created_datetime = oci_par.time_created.replace(
                                tzinfo=_datetime.timezone.utc)

        expires_datetime = oci_par.time_expires.replace(
                                tzinfo=_datetime.timezone.utc)

        # the URI returned by OCI does not include the server. We need
        # to get the server based on the region of this bucket
        url = _get_object_url_for_region(bucket["region"],
                                         oci_par.access_uri)

        # get the checksum for this URL - used to validate the close
        # request
        from Acquire.ObjectStore import OSPar as _OSPar
        from Acquire.ObjectStore import OSParRegistry as _OSParRegistry
        url_checksum = _OSPar.checksum(url)

        driver_details = {"driver": "oci",
                          "bucket": bucket["bucket_name"],
                          "created_datetime": created_datetime,
                          "par_id": oci_par.id,
                          "par_name": oci_par.name}

        par = _OSPar(url=url, encrypt_key=encrypt_key,
                     key=oci_par.object_name,
                     expires_datetime=expires_datetime,
                     is_readable=readable,
                     is_writeable=writeable,
                     driver_details=driver_details)

        _OSParRegistry.register(par=par,
                                url_checksum=url_checksum,
                                details_function=_get_driver_details_from_par,
                                cleanup_function=cleanup_function)

        return par

    @staticmethod
    def close_par(par=None, par_uid=None, url_checksum=None):
        """Close the passed OSPar, which provides access to data in the
           passed bucket

           Args:
                par (OSPar, default=None): OSPar to close bucket
                par_uid (str, default=None): UID for OSPar
                url_checksum (str, default=None): Checksum to
                pass to PARRegistry
           Returns:
                None
        """
        from Acquire.ObjectStore import OSParRegistry as _OSParRegistry

        if par is None:
            par = _OSParRegistry.get(
                            par_uid=par_uid,
                            details_function=_get_driver_details_from_data,
                            url_checksum=url_checksum)

        from Acquire.ObjectStore import OSPar as _OSPar
        if not isinstance(par, _OSPar):
            raise TypeError("The OSPar must be of type OSPar")

        if par.driver() != "oci":
            raise ValueError("Cannot delete a OSPar that was not created "
                             "by the OCI object store")

        # delete the PAR
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        par_bucket = par.driver_details()["bucket"]
        par_id = par.driver_details()["par_id"]

        bucket = _get_service_account_bucket()

        # now get the bucket accessed by the OSPar...
        bucket = OCI_ObjectStore.get_bucket(bucket=bucket,
                                            bucket_name=par_bucket)

        client = bucket["client"]

        try:
            response = client.delete_preauthenticated_request(
                                            client.get_namespace().data,
                                            bucket["bucket_name"],
                                            par_id)
        except Exception as e:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to delete a OSPar '%s' : Error %s" %
                (par_id, str(e)))

        if response.status not in [200, 204]:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "Unable to delete a OSPar '%s' : Status %s, Error %s" %
                (par_id, response.status, str(response.data)))

        # close the OSPar - this will trigger any close_function(s)
        _OSParRegistry.close(par=par)

    @staticmethod
    def get_object(bucket, key):
        """Return the binary data contained in the key 'key' in the
           passed bucket

           Args:
                bucket (dict): Bucket containing data
                key (str): Key for data in bucket
           Returns:
                bytes: Binary data

        """

        key = _clean_key(key)

        try:
            response = bucket["client"].get_object(bucket["namespace"],
                                                   bucket["bucket_name"],
                                                   key)
            is_chunked = False
        except:
            try:
                response = bucket["client"].get_object(bucket["namespace"],
                                                       bucket["bucket_name"],
                                                       "%s/1" % key)
                is_chunked = True
            except:
                is_chunked = False
                pass

            # Raise the original error
            if not is_chunked:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError("No data at key '%s'" % key)

        data = None

        for chunk in response.data.raw.stream(1024 * 1024,
                                              decode_content=False):
            if not data:
                data = chunk
            else:
                data += chunk

        if is_chunked:
            # keep going through to find more chunks
            next_chunk = 1

            while True:
                next_chunk += 1

                try:
                    response = bucket["client"].get_object(
                                        bucket["namespace"],
                                        bucket["bucket_name"],
                                        "%s/%d" % (key, next_chunk))
                except:
                    response = None
                    break

                for chunk in response.data.raw.stream(1024 * 1024,
                                                      decode_content=False):
                    if not data:
                        data = chunk
                    else:
                        data += chunk

        return data

    @staticmethod
    def take_object(bucket, key):
        """Take (delete) the object from the object store, returning
           the object

           Args:
                bucket (dict): Bucket containing data
                key (str): Key for data

           Returns:
                bytes: Binary data
        """
        # ideally the get and delete should be atomic... would like this API
        data = OCI_ObjectStore.get_object(bucket, key)

        try:
            OCI_ObjectStore.delete_object(bucket, key)
        except:
            pass

        return data

    @staticmethod
    def get_all_object_names(bucket, prefix=None, without_prefix=False):
        """Returns the names of all objects in the passed bucket

           Args:
                bucket (dict): Bucket containing data
                prefix (str): Prefix for data
           Returns:
                list: List of all objects in bucket

        """
        if prefix is not None:
            prefix = _clean_key(prefix)

        objects = bucket["client"].list_objects(bucket["namespace"],
                                                bucket["bucket_name"],
                                                prefix=prefix).data

        names = []

        if without_prefix:
            prefix_len = len(prefix)

        for obj in objects.objects:
            if prefix:
                if obj.name.startswith(prefix):
                    name = obj.name
            else:
                name = obj.name

            while name.endswith("/"):
                name = name[0:-1]

            while name.startswith("/"):
                name = name[1:]

            if without_prefix:
                name = name[prefix_len:]

                while name.startswith("/"):
                    name = name[1:]

            if len(name) > 0:
                names.append(name)

        return names

    @staticmethod
    def set_object(bucket, key, data):
        """Set the value of 'key' in 'bucket' to binary 'data'

           Args:
                bucket (dict): Bucket containing data
                key (str): Key for data in bucket
                data (bytes): Binary data to store in bucket

           Returns:
                None
        """
        if data is None:
            data = b'0'

        f = _io.BytesIO(data)

        key = _clean_key(key)
        bucket["client"].put_object(bucket["namespace"],
                                    bucket["bucket_name"],
                                    key, f)

    @staticmethod
    def delete_all_objects(bucket, prefix=None):
        """Deletes all objects...

           Args:
                bucket (dict): Bucket containing data
                prefix (str, default=None): Prefix for data,
                currently unused
            Returns:
                None
        """

        for obj in OCI_ObjectStore.get_all_object_names(bucket):
            bucket["client"].delete_object(bucket["namespace"],
                                           bucket["bucket_name"],
                                           obj)

    @staticmethod
    def delete_object(bucket, key):
        """Removes the object at 'key'

           Args:
                bucket (dict): Bucket containing data
                key (str): Key for data
           Returns:
                None
        """
        try:
            key = _clean_key(key)
            bucket["client"].delete_object(bucket["namespace"],
                                           bucket["bucket_name"],
                                           key)
        except:
            pass

    @staticmethod
    def get_size_and_checksum(bucket, key):
        """Return the object size (in bytes) and MD5 checksum of the
           object in the passed bucket at the specified key

           Args:
                bucket (dict): Bucket containing data
                key (str): Key for object
           Returns:
                tuple (int, str): Size and MD5 checksum of object

        """
        key = _clean_key(key)

        try:
            response = bucket["client"].get_object(bucket["namespace"],
                                                   bucket["bucket_name"],
                                                   key)
        except:
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError("No data at key '%s'" % key)

        content_length = response.headers["Content-Length"]
        checksum = response.headers["Content-MD5"]

        # the checksum is a base64 encoded Content-MD5 header
        # described as standard part of HTTP RFC 2616. Need to
        # convert this back to a hexdigest
        import binascii as _binascii
        import base64 as _base64
        md5sum = _binascii.hexlify(_base64.b64decode(checksum)).decode("utf-8")

        return (int(content_length), md5sum)




from uuid import uuid4 as _uuid4
import copy as _copy
import os as _os

from Acquire.Access import Request as _Request

__all__ = ["FileWriteRequest"]


def _get_key(root_key, filename):
    """Return the object store key for 'filename', using 'root_key' as
       the root

        Args:
            root_key (str): root_key for accessing object store
            filename (str): filename to access in object store
        Returns:
            str: Location to access the file
    """
    basename = _os.path.basename(filename)

    if root_key:
        return "%s/%s" % (str(root_key), basename)
    else:
        return basename


def _clean_key(root_key, filekey):
    """Return the cleaned key 'filekey', using 'root_key' as the root

      Args:
            root_key (str): root_key for accessing object store
            filekey (str): filename to access in object store
        Returns:
            str: Location to access the file

    """
    if root_key:
        return "%s/%s" % (str(root_key), str(filekey))
    else:
        return str(filekey)


def _get_filesize_and_checksum(filename):
    """Calculates the size in bytes of the file and
        the MD5 hash

       Args:
            filename (str): name of file to calculate hash for
        Returns:
            tuple (int, str): size of file in bytes, md5 hash of file
    """
    from hashlib import md5 as _md5
    md5 = _md5()
    size = 0

    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5.update(chunk)
            size += len(chunk)

    return (size, str(md5.hexdigest()))


def _list_all_files(directory, ignore_hidden=True):
    """Create a list of the path relative to 'directory' of
       all files contained in 'directory'.

       Args:
            directory (str): path of directory to traverse
            ignore_hidden (bool): Default=True, if True then include all
            hidden files, else ignore hidden files

        Returns:
            list: a list of all files in the directory

    """

    if not _os.path.isdir(directory):
        return []

    all_files = []

    for (root, _, filenames) in _os.walk(directory):
        root = root[len(directory)+1:]

        if ignore_hidden and root.startswith("."):
            continue

        for filename in filenames:
            if not (filename.startswith(".") and ignore_hidden):
                all_files.append(_os.path.join(root, filename))

    return all_files


def _clean_path(path):
    """This function cleans the passed path so that it doesn't contain
       redundant slashes or '..' etc., so that all backslashes are forwards
       slashes, and that the trailing slash is removed

       Args:
            path (str): the path string to clean
        Returns:
            str: the path string cleaned of extra characters

    """

    if path is None:
        return ""

    path = _os.path.normpath(path)

    # remove all ".." and "." from this path
    if path.find(".") != -1:
        parts = path.split("/")
        for i, part in enumerate(parts):
            if part == "." or part == "..":
                parts[i] = ""

        path = _os.path.normpath("/".join(parts))

    return path


def _expand_source_destination(source, destination=None,
                               root=None, ignore_hidden=True):
    """This function expands the 'source' and 'destination' into a pair
       of lists - the source files and the destination keys in the
       object store.

       Args:
            source (str): source directory to traverse
            destination (str, optional): destination directory
            root (str): root key to use in object store
            ignore_hidden (bool): If True ignore hidden files in folders, else
            include hidden files

        Returns:
            tuple (list, list): lists of files in the source and destination
            directories

    """

    if source is None:
        return ([], [])

    if not isinstance(source, list):
        source = [str(source)]

    is_destination_dir = False
    try:
        is_destination_dir = destination.endswith("/")
    except:
        pass

    destination = _os.path.normpath(_os.path.join(_clean_path(root),
                                                  _clean_path(destination)))

    # expand all of the sources into the full paths of files (which must exist)
    source_filenames = []
    rel_filenames = []

    from glob import glob as _glob

    for s in source:
        for f in _glob(str(s)):
            abspath = _os.path.abspath(f)
            if not _os.path.exists(abspath):
                raise FileExistsError("The file '%s' does not exist!"
                                      % abspath)

            if _os.path.isdir(abspath):
                dirfiles = _list_all_files(abspath, ignore_hidden)

                for dirfile in dirfiles:
                    source_filenames.append(_os.path.join(abspath, dirfile))
                    rel_filenames.append(dirfile)
            else:
                source_filenames.append(abspath)
                rel_filenames.append(_os.path.basename(abspath))

    destination_keys = []

    if len(rel_filenames) == 1:
        if is_destination_dir:
            destination_keys.append(_os.path.join(destination,
                                                  rel_filenames[0]))
        else:
            destination_keys.append(destination)
    else:
        for rel_filename in rel_filenames:
            destination_keys.append(_os.path.join(destination, rel_filename))

    if len(source_filenames) != len(destination_keys):
        raise ValueError("Something went wrong as the number of source "
                         "filenames should be equal to the number of keys")

    return (source_filenames, destination_keys)


class FileWriteRequest(_Request):
    """This class holds a request to write a file (or files) to a remote
       object store
    """
    def __init__(self, source=None, destination=None, root=None,
                 ignore_hidden=True, account=None, testing_key=None):
        """Construct the request to write the files specified in 'source'
           to 'destination' in the cloud object store. You can optionally
           pass a 'root' for all of the keys in the object store, and,
           if the source is a directory, you can ignore hidden files using
           'ignore_hidden=True'.

           You must pass the 'account' from which payment will be taken to
           write files to the object store.

            Args:
                source (str): source directory to traverse
                destination (str, optional): destination directory
                root (str): root key to use in object store
                ignore_hidden (bool): If True ignore hidden files in folders,
                else include hidden files
                account (Account): instance of Account class
                testing_key (str): passed to enable testing of class
            Returns:
                None

        """
        super().__init__()

        (filenames, keys) = FileWriteRequest.expand_source_and_destination(
                                                source, destination,
                                                root, ignore_hidden)

        self._destination_keys = keys
        self._source_filenames = filenames
        self._file_sizes = []
        self._checksums = []

        if len(self._source_filenames) == 0:
            self._uid = None
            return

        for filename in self._source_filenames:
            (size, checksum) = _get_filesize_and_checksum(filename)
            self._file_sizes.append(size)
            self._checksums.append(checksum)

        self._is_testing = False

        if testing_key:
            self._is_testing = True
            self._testing_key = testing_key
        elif account is not None:
            from Acquire.Client import Account as _Account
            if not isinstance(account, _Account):
                raise TypeError("The passed account must be type Account")

            if not account.is_logged_in():
                raise PermissionError(
                    "You can only authorise payment from the account if you "
                    "have logged in")

        else:
            raise ValueError("You must pass a valid account to write files!")

        self._uid = str(_uuid4())

        from Acquire.Identity import Authorisation as _Authorisation

        if self._is_testing:
            self._account_uid = "something"
            self._accounting_service_url = "somewhere"
            self._authorisation = _Authorisation(
                                    resource=self.resource_key(),
                                    testing_key=self._testing_key)

        else:
            self._account_uid = account.uid()
            self._accounting_service_url = account.accounting_service() \
                                                  .canonical_url()

            self._authorisation = _Authorisation(
                                    resource=self.resource_key(),
                                    user=account.owner())

    def is_null(self):
        """Return whether or not this is a null request

            Returns:
                bool: True if UID set, else False
        """
        return self._uid is None

    def __str__(self):
        if self.is_null():
            return "FileWriteRequest::null"
        else:
            return "FileWriteRequest(uid=%s)" % self._uid

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def resource_key(self):
        """Function to return a string that can be used as a
           summary key for this resource request


            Returns:
                None or str : None if instance is null, else string containing
                uid and checksum
        """
        if self.is_null():
            return None
        else:
            return "%s %s" % (self._uid, ":".join(self._checksums))

    def uid(self):
        """Get the UID of this request

        Returns:
            str: UID of this request

        """
        return self._uid

    def authorisation(self):
        """Return the authorisation behind this request

            Returns:
                Authorisation: the authorisation behind this request
        """
        return self._authorisation

    def source_filenames(self):
        """Return the filenames of the files to be copied. Note that this
           information is only available in the copy of the object that
           created the request - it is not saved when this object is
           serialised to json as we don't want to leak potentially
           sensitive data to the object store

           Returns:
                list: list of files to be copied
        """
        return self._source_filenames

    def destination_keys(self):
        """Return the object store keys to which the files will be
           written

           Returns:
                list: object store keys to which the files will be written

        """
        return _copy.copy(self._destination_keys)

    def filesizes(self):
        """Return the sizes of the files that are requested to be written

            Returns:
                list: size of the files to be written

        """
        return _copy.copy(self._file_sizes)

    def checksums(self):
        """Return the checksums of the files that are requested
           to be written

           Returns:
                list: checksums of the files to be written

           """
        return _copy.copy(self._checksums)

    def account_uid(self):
        """Return the UID of the account from which payment should be
           taken for the file storage

           Returns:
                str or None: returns the UID of the account if valid
                else None
        """
        try:
            return self._account_uid
        except:
            return None

    @staticmethod
    def expand_source_and_destination(source, destination, root=None,
                                      ignore_hidden=True):
        """This function expands the passed source and destination objects
           into the list of absolute paths of local source files, and the
           full keys of those files once they are uploaded to the object
           store. This returns a pair of lists - the lists match the
           absolute path of the local file to the desired full key
           of the file in the object store

           Args:
                source (list): list of source objects
                destination (list): list of destinations for objects
                root (str, optional, default=None): root key for objects
                ignore_hidden (bool, optional, default=True): ignore
                hidden files
        """
        return _expand_source_destination(source, destination, root,
                                          ignore_hidden)

    def accounting_service_url(self):
        """Return the canonical URL of the service holding the account

            Returns:
                str or None: If _accounting_service_url valid return
                return it else return None
        """
        try:
            return self._accounting_service_url
        except:
            return None

    def to_data(self):
        """Return this request as a json-serialisable dictionary

            Returns:
                dict: a JSON-serialisable dictionary of this request

        """
        if self.is_null():
            return {}

        data = super().to_data()
        data["uid"] = self._uid
        data["destination_keys"] = self._destination_keys
        data["file_sizes"] = self._file_sizes
        data["checksums"] = self._checksums
        data["authorisation"] = self._authorisation.to_data()
        data["account_uid"] = self._account_uid
        data["accounting_service_url"] = self._accounting_service_url

        # don't write self._source_filenames as this contains
        # potentially sensitive local data that is not needed
        # by the cloud object store

        return data

    @staticmethod
    def from_data(data):
        """ Create a request from a JSON object

            Args:
                data (str) : an object serialised as a JSON object
            Returns:
                FileWriteRequest: a FileWriteRequest object created from
                the JSON data

        """
        if (data and len(data) > 0):
            from Acquire.Identity import Authorisation as _Authorisation
            f = FileWriteRequest()

            f._uid = data["uid"]
            f._destination_keys = data["destination_keys"]
            f._file_sizes = data["file_sizes"]
            f._checksums = data["checksums"]
            f._authorisation = _Authorisation.from_data(data["authorisation"])
            f._account_uid = data["account_uid"]
            f._accounting_service_url = data["accounting_service_url"]

            f._from_data(data)

            return f
        else:
            return None




__all__ = ["Accounts"]


class Accounts:
    """This class provides the interface to grouping and ungrouping
       accounts, and associating them with users and services. An account
       can belong to multiple groups. Authorisation of access to accounts
       is based on which group they are in. Typically the groups will
       refer to users, e.g. all of the accounts for a user will be in the
       user's group, and can then be authorised by authorising the user
    """
    def __init__(self, user_guid=None, group=None, aclrules=None):
        """Construct an interface to the group of accounts called 'group'.
           If the group is not specified, then it will default to 'default'

           The passed ACLRules are used to control access to this group.
           Note that these control access to the group.
        """
        if group is None:
            if user_guid is None:
                group = "default"
            else:
                group = user_guid

        self._group = str(group)
        self._get_aclrules(user_guid=user_guid, aclrules=aclrules)
        self._user_guid = user_guid

    def __str__(self):
        return "Accounts(group=%s)" % self.group()

    def _root(self):
        """Return the root key for this group in the object store

            Returns:
                string: Root key for group
        """
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded
        return "accounting/account_groups/%s" % \
            _string_to_encoded(self._group)

    def _get_aclrules(self, user_guid, aclrules, bucket=None):
        """Load up the ACLRules for this group. If none are set, then
           either the passed ACLRules will be used, or the specified
           user will be set as the owner
        """
        from Acquire.Identity import ACLRules as _ACLRules
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        aclkey = self._acls_key()

        try:
            self._aclrules = _ACLRules.from_data(
                                _ObjectStore.get_object_from_json(
                                    bucket=bucket, key=aclkey))
        except:
            self._aclrules = None

        if self._aclrules is not None:
            return

        if aclrules is None:
            if user_guid is None:
                raise PermissionError(
                    "You must specify the guid of the initial user who "
                    "owns this account!")

            aclrules = _ACLRules.owner(user_guid=user_guid)
        elif not isinstance(aclrules, _ACLRules):
            raise TypeError("The ACLRules must be type ACLRules")

        _ObjectStore.set_object_from_json(bucket=bucket, key=aclkey,
                                          data=aclrules.to_data())

        self._aclrules = aclrules

    def identifiers(self):
        """Return the set of validated identifiers that are known to
           be attached to this group of accounts
        """
        return {"user_guid": self._user_guid}

    def aclrules(self):
        """Return the ACL rules for this group of accounts"""
        return self._aclrules

    def _account_key(self, name):
        """Return the key for the account called 'name' in this group

            Args:
                name (str): Name of the account

            Returns:
                str: Key for the account 'name'
        """
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded
        return "%s/%s" % (self._root(),
                          _string_to_encoded(str(name)))

    def _acls_key(self):
        """Return the key for the ACLs for this accounts group"""
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded
        return "accounting/account_group_acls/%s" % \
            _string_to_encoded(self._group)

    def group(self):
        """Return the name of the group that this set of accounts refers to

            Returns:
                string: Name of the group
        """
        return self._group

    def name(self):
        """Return the name of the group that this set of accounts refers to"""
        return self._group

    def _assert_is_readable(self):
        """Assert that we have permission to read these accounts"""
        aclrule = self._aclrules.resolve(must_resolve=True,
                                         identifiers=self.identifiers())

        if not aclrule.is_readable():
            raise PermissionError(
                "You do not have permission to access the accounts "
                "in this group")

    def _assert_is_writeable(self):
        """Assert that we have permission to write these accounts"""
        aclrule = self._aclrules.resolve(must_resolve=True,
                                         identifiers=self.identifiers())

        if not aclrule.is_writeable():
            raise PermissionError(
                "You do not have permission to write to the accounts "
                "in this group")

    def list_accounts(self, bucket=None):
        """Return the names of all of the accounts in this group

            Args:
                bucket (dict, default=None): Bucket from which to load data

            Returns:
                :obj:`list`: List of names of the accounts in this group

        """
        self._assert_is_readable()

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import encoded_to_string \
            as _encoded_to_string

        keys = _ObjectStore.get_all_object_names(bucket, self._root())
        root_len = len(self._root())

        accounts = []

        for key in keys:
            try:
                account_key = key[root_len:]
                while account_key.startswith("/"):
                    account_key = account_key[1:]

                accounts.append(_encoded_to_string(account_key))
            except Exception as e:
                from Acquire.Accounting import AccountError
                raise AccountError(
                    "Unable to identify the account associated with key "
                    "'%s', equals '%s': %s" %
                    (key, account_key, str(e)))

        return accounts

    def get_account(self, name, bucket=None):
        """Return the account called 'name' from this group

            Args:
                name (:obj:`str`): Name of account to retrieve
                bucket (:obj:`dict`): Bucket to load data from

            Returns:
                :obj:`Account`: Account object
        """
        self._assert_is_readable()

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        try:
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            account_uid = _ObjectStore.get_string_object(
                            bucket, self._account_key(name))
        except:
            account_uid = None

        if account_uid is None:
            # ensure that the user always has a "main" account
            if name == "main":
                return self.create_account("main", "primary user account",
                                           overdraft_limit=0, bucket=bucket)

            from Acquire.Accounting import AccountError
            raise AccountError("There is no account called '%s' in the "
                               "group '%s'" % (name, self.group()))

        from Acquire.Accounting import Account as _Account
        return _Account(uid=account_uid, bucket=bucket)

    def contains(self, account, bucket=None):
        """Return whether or not this group contains the passed account

            Args:
                account (:obj:`Account`): Account to check against group
                bucket (dict, default=None): Bucket to load data from
            Returns:
                bool : True if account in group, else False
        """
        self._assert_is_readable()

        from Acquire.Accounting import Account as _Account
        if not isinstance(account, _Account):
            raise TypeError("The passed account must be of type Account")

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        # read the UID of the account in this group that matches the
        # passed account's name
        try:
            from Acquire.ObjectStore import ObjectStore as _ObjectStore
            account_uid = _ObjectStore.get_string_object(
                            bucket, self._account_key(account.name()))
        except:
            account_uid = None

        return account.uid() == account_uid

    def create_account(self, name, description=None,
                       overdraft_limit=None, bucket=None,
                       authorisation=None):
        """Create a new account called 'name' in this group. This will
           return the existing account if it already exists

           Args:
                name (str): Name of account to create
                description (default=None): Description of account
                overdraft_limit (int, default=None): Limit of overdraft
                bucket (dict, default=None): Bucket to load data from

            Returns:
                Account: New Account object
        """
        if name is None:
            raise ValueError("You must pass a name of the new account")

        from Acquire.Identity import Authorisation as _Authorisation

        if authorisation is not None:
            if not isinstance(authorisation, _Authorisation):
                raise TypeError("The authorisation must be type Authorisation")

            if self._user_guid is not None:
                if self._user_guid != authorisation.user_guid():
                    raise PermissionError(
                        "The same user who opened this accounts group (%s) "
                        "must create accounts in this group (%s)" %
                        (self._user_guid, authorisation.user_guid()))

            authorisation.verify("create_account %s" % name)

        self._assert_is_writeable()

        account_key = self._account_key(name)

        if bucket is None:
            from Acquire.Service import get_service_account_bucket \
                as _get_service_account_bucket
            bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Accounting import Account as _Account

        try:
            account_uid = _ObjectStore.get_string_object(bucket, account_key)
        except:
            account_uid = None

        if account_uid is not None:
            # this account already exists - just return it
            account = _Account(uid=account_uid, bucket=bucket)

            if account.group_name() != self.name():
                account.set_group(self)

            if overdraft_limit is not None:
                account.set_overdraft_limit(overdraft_limit, bucket=bucket)

            return account

        # make sure that no-one has created this account before
        from Acquire.ObjectStore import Mutex as _Mutex
        m = _Mutex(account_key, timeout=600, lease_time=600, bucket=bucket)

        try:
            account_uid = _ObjectStore.get_string_object(bucket, account_key)
        except:
            account_uid = None

        if account_uid is not None:
            m.unlock()
            # this account already exists - just return it
            account = _Account(uid=account_uid, bucket=bucket)

            if account.group_name() != self.name():
                account.set_group(self)

            if overdraft_limit is not None:
                account.set_overdraft_limit(overdraft_limit, bucket=bucket)

            return account

        # write a temporary UID to the object store so that we
        # can ensure we are the only function to create it
        try:
            _ObjectStore.set_string_object(bucket, account_key,
                                           "under_construction")
        except:
            m.unlock()
            raise

        m.unlock()

        # ok - we are the only function creating this account. Let's try
        # to create it properly. The account is created with the same
        # ACLRules of the group.
        try:
            from Acquire.Identity import ACLRules as _ACLRules
            account = _Account(name=name, description=description,
                               group_name=self.name(),
                               aclrules=_ACLRules.inherit(),
                               bucket=bucket)
        except:
            try:
                _ObjectStore.delete_object(bucket, account_key)
            except:
                pass

            raise

        if overdraft_limit is not None:
            account.set_overdraft_limit(overdraft_limit, bucket=bucket)

        _ObjectStore.set_string_object(bucket, account_key, account.uid())

        return account




_par_root = "storage/par"


class PARRegistry:
    """This object is used by the storage service to manage
       a registry of all of the active PARs in the service
    """
    def __init__(self):
        pass

    def register(self, par, authorisation, secret=None):
        """Register the passed par, which is authorised using the
           passed authorisation. If the authorisation is correct
           this this will return the URL of the PAR
        """
        from Acquire.Client import PAR as _PAR
        from Acquire.Client import Authorisation as _Authorisation

        if not isinstance(par, _PAR):
            raise TypeError("The par must be type PAR")

        # create a new UID for this PAR
        from Acquire.ObjectStore import create_uid as _create_uid
        uid = _create_uid()
        par._set_uid(uid)

        if par.expired():
            raise PermissionError("The passed PAR has already expired!")

        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The authorisation must be type Authorisation")

        identifiers = authorisation.verify(
                            resource="create_par %s" % par.fingerprint(),
                            return_identifiers=True)

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        if secret is not None and len(secret) > 0:
            from Acquire.Crypto import Hash
            secret = Hash.multi_md5(uid, secret)
        else:
            secret = None

        import json as _json
        data = {"par": par.to_data(),
                "identifiers": _json.dumps(identifiers),
                "secret": secret}

        key = "%s/%s" % (_par_root, uid)

        bucket = _get_service_account_bucket()
        _ObjectStore.set_object_from_json(bucket, key, data)

        return uid

    def load(self, par_uid, secret=None):
        """Load and return the PAR and identifiers associated with
           the passed UID, locked with the passed secret
        """
        # validate that the UID actually looks like a UID. This
        # should prevent attacks that try weird UIDs
        from Acquire.ObjectStore import validate_is_uid \
            as _validate_is_uid
        _validate_is_uid(par_uid)

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        try:
            key = "%s/%s" % (_par_root, par_uid)
            bucket = _get_service_account_bucket()
            data = _ObjectStore.get_object_from_json(bucket, key)

            from Acquire.Client import PAR as _PAR
            import json as _json
            par = _PAR.from_data(data["par"])
            identifiers = _json.loads(data["identifiers"])

            if secret != data["secret"]:
                raise PermissionError()
        except:
            raise PermissionError(
                "There is no valid PAR at ID '%s'" % par_uid)

        if par.expired():
            raise PermissionError(
                "There is no valid PAR at ID '%s' as it has expired" % par_uid)

        return (par, identifiers)

    def resolve(self, par_uid, secret=None):
        """Resolve the passed PAR and return the corresponding
           DriveMeta or FileMeta with requisite permissions
        """
        (par, identifiers) = self.load(par_uid, secret)

        from Acquire.Storage import DriveInfo as _DriveInfo
        from Acquire.Storage import DriveMeta as _DriveMeta

        # Load up the drive with specified drive_uid so that we
        # can inspect the aclrules that control access
        drive = _DriveInfo(drive_uid=par.location().drive_uid())

        drivemeta = _DriveMeta(name="par:%s" % par.uid(),
                               uid=drive.uid(),
                               aclrules=drive.aclrules())

        # now resolve the aclrule for the user who created the
        # PAR, and that no more permissions are granted than those
        # specified in the PAR
        drive_acl = drivemeta.resolve_acl(identifiers=identifiers,
                                          open_aclrule=par.aclrule())

        if drive_acl.denied_all():
            raise PermissionError(
                "The PAR does not have permission to access the drive")

        if par.location().is_drive():
            return drivemeta

        filemetas = drive.list_files(par=par, identifiers=identifiers,
                                     include_metadata=True)

        return filemetas





__all__ = ["ACLRule"]


def _max(a, b):
    """Function to do most-permissive addition of two bools

        True + True = True
        True + False = True
        False + True = True
        False + False = False

        True + None = True
        None + True = True

        False + None = False
        None + False = False

        None + None = None
    """
    if a is None:
        return b
    elif b is None:
        return a
    elif a is True:
        return True
    else:
        return b


def _sub(a, b):
    """Function to do subtraction of two bools

        True - True = False
        True - False = True
        False - True = False
        False - False = False

        True - None = True
        None - True = False

        False - None = False
        None - False = None

        None - None = None
    """
    if b is True:
        return False
    else:
        return a


def _min(a, b):
    """Function to do least-permission addition of two bools

        True * True = True
        True * False = False
        False * True = False
        False * False = False

        True * None = True
        None * True = True

        False * None = False
        None * False = False

        None - None = None
    """
    if a is None:
        return b
    elif b is None:
        return a
    elif a is False:
        return False
    else:
        return b


class ACLRule:
    """This class holds the access control list (ACL) rule for
       a particular user accessing a particular resource
    """
    def __init__(self, is_owner=None, is_readable=None,
                 is_writeable=None, is_executable=None):
        """Construct a default rule. By default this rule has
           fully-inherit permissions (nothing is set either way)
        """
        if is_owner is None:
            self._is_owner = None
        elif is_owner:
            self._is_owner = True
        else:
            self._is_owner = False

        if is_readable is None:
            self._is_readable = None
        elif is_readable:
            self._is_readable = True
        else:
            self._is_readable = False

        if is_writeable is None:
            self._is_writeable = None
        elif is_writeable:
            self._is_writeable = True
        else:
            self._is_writeable = False

        if is_executable is None:
            self._is_executable = None
        elif is_executable:
            self._is_executable = True
        else:
            self._is_executable = False

    def __str__(self):
        s = []

        if self.is_owner():
            s.append("owner")
        elif self.inherits_owner():
            s.append("inherits_owner")

        if self.is_writeable():
            s.append("writeable")
        elif self.inherits_writeable():
            s.append("inherits_writeable")

        if self.is_readable():
            s.append("readable")
        elif self.inherits_readable():
            s.append("inherits_readable")

        if self.is_executable():
            s.append("executable")
        elif self.inherits_executable():
            s.append("inherits_executable")

        if len(s) == 0:
            return "ACLRule::denied"
        else:
            return "ACLRule(%s)" % ", ".join(s)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __add__(self, other):
        """Add two rules together - this will combine the parts
           additively, i.e. the most-permissive options are set
           in the return value
        """
        if not isinstance(other, ACLRule):
            raise TypeError(
                "You can only combine together pairs of ACLRules")

        return ACLRule(
                is_owner=_max(self._is_owner, other._is_owner),
                is_writeable=_max(self._is_writeable, other._is_writeable),
                is_readable=_max(self._is_readable, other._is_readable),
                is_executable=_max(self._is_executable, other._is_executable))

    def __sub__(self, other):
        """Subtract 'other' from these rules - if 'other' is False
           the this is set to False
        """
        if not isinstance(other, ACLRule):
            raise TypeError(
                "You can only combine together pairs of ACLRules")

        return ACLRule(
                is_owner=_sub(self._is_owner, other._is_owner),
                is_writeable=_sub(self._is_writeable, other._is_writeable),
                is_readable=_sub(self._is_readable, other._is_readable),
                is_executable=_sub(self._is_executable, other._is_executable))

    def __mul__(self, other):
        """Combine two rules together - this will combine
           the parts subtractively, i.e. the least permissive options
           are set in the return value
        """
        if not isinstance(other, ACLRule):
            raise TypeError(
                "You can only combine together pairs of ACLRules")

        return ACLRule(
                is_owner=_min(self._is_owner, other._is_owner),
                is_writeable=_min(self._is_writeable, other._is_writeable),
                is_readable=_min(self._is_readable, other._is_readable),
                is_executable=_min(self._is_executable, other._is_executable))

    @staticmethod
    def owner():
        """Return the ACLRule of an owner"""
        return ACLRule(is_owner=True, is_readable=True,
                       is_writeable=True, is_executable=True)

    @staticmethod
    def executer():
        """Return the ACLRule of an executer"""
        return ACLRule(is_owner=False, is_executable=True,
                       is_writeable=True, is_readable=True)

    @staticmethod
    def writer():
        """Return the ACLRule of a writer"""
        return ACLRule(is_owner=False, is_readable=True,
                       is_writeable=True, is_executable=False)

    @staticmethod
    def reader():
        """Return the ACLRule of a reader"""
        return ACLRule(is_owner=False, is_readable=True,
                       is_writeable=False, is_executable=False)

    @staticmethod
    def denied():
        """Return a "denied" (no-access at all) permission rule"""
        return ACLRule(is_owner=False, is_readable=False,
                       is_writeable=False, is_executable=False)

    @staticmethod
    def null():
        """Return a null (inherit all permissions) rule"""
        return ACLRule(is_owner=None, is_readable=None,
                       is_writeable=None, is_executable=None)

    @staticmethod
    def inherit():
        """Return the ACL rule that means 'inherit permissions from parent'"""
        return ACLRule(is_owner=None, is_readable=None,
                       is_writeable=None, is_executable=None)

    def fingerprint(self):
        """A fingerprint that can be used to validate a request"""
        return "%d.%d.%d.%d" % (self._is_owner, self._is_executable,
                                self._is_writeable, self._is_readable)

    def is_owner(self):
        """Return whether or not the user is the owner of this resource"""
        return self._is_owner

    def is_readable(self):
        """Return whether or not the user can read this resource"""
        return self._is_readable

    def is_writeable(self):
        """Return whether or not the user can write to this resource"""
        return self._is_writeable

    def is_executable(self):
        """Return whether or not the user can execute this resource"""
        return self._is_executable

    def is_denied(self):
        """Return whether or not this rule is all denied"""
        return self.denied_all()

    def inherits_owner(self):
        """Return whether or not this inherits the owner status
           from upstream
        """
        return self._is_owner is None

    def inherits_readable(self):
        """Return whether or not this inherits the reader status
           from upstream
        """
        return self._is_readable is None

    def inherits_writeable(self):
        """Return whether or not this inherits the writeable status
           from upstream
        """
        return self._is_writeable is None

    def inherits_executable(self):
        """Return whether or not this inherits the executable status
           from upstream
        """
        return self._is_executable is None

    def inherits_all(self):
        """Return whether or not this rule inherits all permissions"""
        return (self._is_owner is None) and \
               (self._is_readable is None) and \
               (self._is_writeable is None) and \
               (self._is_executable is None)

    def is_fully_resolved(self):
        """Return whether or not this rule is fully resolved"""
        return (self._is_owner is not None) and \
               (self._is_readable is not None) and \
               (self._is_writeable is not None) and \
               (self._is_executable is not None)

    def denied_all(self):
        """Return whether or not this rule shows that everything is
           denied
        """
        return (self._is_owner is False) and \
               (self._is_readable is False) and \
               (self._is_writeable is False) and \
               (self._is_executable is False)

    def _force_resolve(self, unresolved=False):
        """This returns a fully resolved ACL, where anything that
           is not resolved is set equal to 'unresolved'
        """
        if unresolved:
            unresolved = True
        else:
            unresolved = False

        is_owner = self._is_owner
        is_readable = self._is_readable
        is_writeable = self._is_writeable
        is_executable = self._is_executable

        if is_owner is None:
            is_owner = unresolved

        if is_readable is None:
            is_readable = unresolved

        if is_executable is None:
            is_executable = unresolved

        if is_writeable is None:
            is_writeable = unresolved

        return ACLRule(is_owner=is_owner, is_writeable=is_writeable,
                       is_readable=is_readable, is_executable=is_executable)

    def resolve(self, must_resolve=True, identifiers=None,
                upstream=None, unresolved=False):
        """Resolve these rules based on the information supplied
           in 'identifiers'. Notably, if any of our rules are 'inherit',
           the this will look into 'upstream' to
           inherit the rule. If 'must_resolve' is true, then
           this function must always return a fully-resolved ACLRule
           (in which any unresolved parts are set equal to 'unresolved')
        """
        if self.is_fully_resolved():
            return self

        if upstream is None:
            if must_resolve:
                return self._force_resolve(unresolved=unresolved)
            else:
                return self

        from Acquire.Identity import ACLRules as _ACLRules
        if isinstance(upstream, _ACLRules):
            upstream = upstream.resolve(must_resolve=False,
                                        identifiers=identifiers)

        if must_resolve and (not upstream.is_fully_resolved()):
            upstream = upstream._force_resolve(unresolved=unresolved)

        is_owner = self._is_owner
        is_readable = self._is_readable
        is_writeable = self._is_writeable
        is_executable = self._is_executable

        if is_owner is None:
            is_owner = upstream._is_owner

        if is_readable is None:
            is_readable = upstream._is_readable

        if is_writeable is None:
            is_writeable = upstream._is_writeable

        if is_executable is None:
            is_executable = upstream._is_executable

        result = ACLRule(is_owner=is_owner, is_writeable=is_writeable,
                         is_readable=is_readable, is_executable=is_executable)

        if must_resolve:
            result = result._force_resolve(unresolved=unresolved)

        return result

    def set_owner(self, is_owner=True):
        """Set the user as an owner of the bucket"""
        if is_owner:
            self._is_owner = True
        else:
            self._is_owner = False

    def set_readable(self, is_readable=True):
        """Set the readable rule to 'is_readable'"""
        if is_readable:
            self._is_readable = True
        else:
            self._is_readable = False

    def set_writeable(self, is_writeable=True):
        """Set the writeable rule to 'is_writeable'"""
        if is_writeable:
            self._is_writeable = True
        else:
            self._is_writeable = False

    def set_inherits_owner(self):
        """Set that this ACL inherits ownership from its parent"""
        self._is_owner = None

    def set_inherits_readable(self):
        """Set that this ACL inherits readable from its parent"""
        self._is_readable = None

    def set_inherits_writeable(self):
        """Set that this ACL inherits writeable from its parent"""
        self._is_writeable = None

    def set_readable_writeable(self, is_readable_writeable=True):
        """Set both the readable and writeable rules to
           'is_readable_writeable'
        """
        if is_readable_writeable:
            self._is_readable = True
            self._is_writeable = True
        else:
            self._is_readable = False
            self._is_writeable = False

    def to_data(self):
        """Return this object converted to a json-serialisable object"""
        if self.inherits_all():
            return "inherits"
        elif self == ACLRule.owner():
            return "owner"
        elif self == ACLRule.reader():
            return "reader"
        elif self == ACLRule.writer():
            return "writer"
        elif self == ACLRule.denied():
            return "denied"
        elif self == ACLRule.executer():
            return "executer"
        else:
            data = {}
            data["is_owner"] = self._is_owner
            data["is_readable"] = self._is_readable
            data["is_writeable"] = self._is_writeable
            data["is_executable"] = self._is_executable
            return data

    @staticmethod
    def from_data(data):
        """Return this object constructed from the passed json-deserialised
           dictionary
        """
        if data is None:
            return None

        if isinstance(data, str):
            if data == "inherits":
                return ACLRule.inherit()
            elif data == "owner":
                return ACLRule.owner()
            elif data == "reader":
                return ACLRule.reader()
            elif data == "writer":
                return ACLRule.writer()
            elif data == "executer":
                return ACLRule.executer()
            elif data == "denied":
                return ACLRule.denied()

        is_owner = None
        is_readable = None
        is_writeable = None

        if "is_owner" in data:
            if data["is_owner"]:
                is_owner = True
            elif data["is_owner"] is not None:
                is_owner = False

        if "is_readable" in data:
            if data["is_readable"]:
                is_readable = True
            elif data["is_readable"] is not None:
                is_readable = False

        if "is_writeable" in data:
            if data["is_writeable"]:
                is_writeable = True
            elif data["is_writeable"] is not None:
                is_writeable = False

        if "is_executable" in data:
            if data["is_executable"]:
                is_executable = True
            elif data["is_executable"] is not None:
                is_executable = False

        return ACLRule(is_owner=is_owner,
                       is_writeable=is_writeable,
                       is_readable=is_readable,
                       is_executable=is_executable)




import uuid as _uuid
from copy import copy as _copy

from Acquire.Service import Service as _Service

__all__ = ["AccountingService"]


class AccountingService(_Service):
    """This is a specialisation of Service for Accounting Services"""
    def __init__(self, other=None):
        if isinstance(other, _Service):
            self.__dict__ = _copy(other.__dict__)

            if not self.is_accounting_service():
                from Acquire.Accounting import AccountingServiceError
                raise AccountingServiceError(
                    "Cannot construct an AccountingService from "
                    "a service which is not an accounting service!")
        else:
            _Service.__init__(self)

    def _call_local_function(self, function, args):
        """Internal function called to short-cut local 'remote'
           function calls

           Args:
                function (function): Function to route
                args: Arguments to pass to routed function

            Returns:
                function: A handler function
        """
        from accounting.route import accounting_functions \
            as _accounting_functions
        from admin.handler import create_handler as _create_handler
        handler = _create_handler(_accounting_functions)
        return handler(function=function, args=args)




__all__ = ["Hash"]


class Hash:
    """This provides a static interface to a range of checksum
       (hashing) functions. By default we use MD5
    """
    @staticmethod
    def md5(data):
        """Return the MD5 checksum of the passed data"""
        if data is None:
            return None

        from hashlib import md5 as _md5

        if isinstance(data, str):
            data = data.encode("utf-8")

        md5 = _md5()
        md5.update(data)
        return md5.hexdigest()

    @staticmethod
    def multi_md5(data1, data2):
        """Return a combined MD5 checksum of data1 and data2 via

           MD5(MD5(data1) + MD5(data2))

           This heavily salts the MD5s of both bits of data
        """
        return Hash.md5(Hash.md5(data1)+Hash.md5(data2))




__all__ = ["Authorisation"]


class Authorisation:
    """This class holds the information needed to show that a user
       has authorised an action. This contains a signed token that
       records the time that the authorisation that was signed, together
       with an extra key (or secret) that can be used by the user
       and provider to verify that the authorisation is for the
       correct resource
    """
    def __init__(self, resource=None, user=None,
                 testing_key=None, testing_user_guid=None):
        """Create an authorisation for the passed resource
           that is authorised by the passed user (who must be authenticated)

           If testing_key is passed, then this authorisation is being
           tested as part of the unit tests
        """

        if resource is not None:
            resource = str(resource)

        self._signature = None
        self._last_validated_datetime = None
        self._scope = None
        self._permissions = None
        self._pubcert = None

        if resource is not None:
            if user is None and testing_key is None:
                raise ValueError(
                    "You must pass in an authenticated user who will "
                    "provide authorisation for resource '%s'" % resource)

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now
        from Acquire.ObjectStore import create_uuid as _create_uuid

        if user is not None:
            from Acquire.Client import User as _User

            if not isinstance(user, _User):
                raise TypeError("The passed user must be of type User")

            elif not user.is_logged_in():
                raise PermissionError(
                    "The passed user '%s' must be authenticated to enable "
                    "you to generate an authorisation for the account")

            self._user_uid = user.uid()
            self._session_uid = user.session_uid()
            self._identity_url = user.identity_service().canonical_url()
            self._identity_uid = user.identity_service_uid()
            self._auth_datetime = _get_datetime_now()
            self._uid = _create_uuid(short_uid=True,
                                     include_date=self._auth_datetime)
            self._siguid = user.signing_key().sign(self._uid)

            message = self._get_message(resource)
            self._signature = user.signing_key().sign(message)

            self._last_validated_datetime = _get_datetime_now()
            self._last_verified_resource = resource
            self._last_verified_key = None

            if user.guid() != self.user_guid():
                # interesting future case when we allow individual users
                # to be identified by multiple identity services...
                raise PermissionError(
                    "We do not yet support a single user being identified "
                    "by multiple identity services: %s versus %s" %
                    (user.guid(), self.user_guid()))

        elif testing_key is not None:
            self._user_uid = "some user uid"
            self._session_uid = "some session uid"
            self._identity_url = "some identity_url"
            self._identity_uid = "some identity uid"
            self._auth_datetime = _get_datetime_now()
            self._uid = _create_uuid(short_uid=True,
                                     include_date=self._auth_datetime)
            self._is_testing = True
            self._testing_key = testing_key

            if testing_user_guid is not None:
                parts = testing_user_guid.split("@")
                self._user_uid = parts[0]
                self._identity_uid = parts[1]

            message = self._get_message(resource)
            self._signature = testing_key.sign(message)
            self._siguid = testing_key.sign(self._uid)

            self._last_validated_datetime = _get_datetime_now()
            self._last_verified_resource = resource
            self._last_verified_key = testing_key.public_key()

    def is_null(self):
        """Return whether or not this authorisation is null"""
        return self._signature is None

    def _get_message(self, resource=None, matched_resource=False):
        """Internal function that is used to generate the message for
           the resource that is signed. This message
           encodes information about the user and identity service that
           signed the message, as well as the resource. This helps
           prevent tamporing with the data in this authorisation.

           If 'matched_resource' is True then this will return the
           message based on the previously-verified resource
           (as we have already determined that the user knows what
           the resource is)
        """
        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string

        if matched_resource:
            resource = self._last_verified_resource

        if resource is None:
            return "%s|%s|%s|%s" % (
                self._user_uid, self._session_uid,
                self._identity_uid,
                _datetime_to_string(self._auth_datetime))
        else:
            return "%s|%s|%s|%s|%s" % (
                self._user_uid, self._session_uid,
                self._identity_uid, str(resource),
                _datetime_to_string(self._auth_datetime))

    def __str__(self):
        try:
            return "Authorisation(signature=%s)" % self._signature
        except:
            return "Authorisation()"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._signature == other._signature
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def _fix_integer(self, value, max_value):
        max_value = int(max_value)

        if value is None:
            return max_value
        else:
            try:
                value = int(value)
            except:
                return max_value

        if value <= 0 or value > max_value:
            return max_value
        else:
            return value

    def from_user(self, user_uid, service_uid):
        """Return whether or not this authorisation comes from the user
           with passed user_uid registered on the passed service_uid
        """
        return (user_uid == self._user_uid) and \
               (service_uid == self._identity_uid)

    def uid(self):
        """Return the UID of this authorisation. This will be signed by
           the user and can be used a use-once record by a service to
           validate that they have not seen this authorisation before
        """
        if self.is_null():
            return None
        else:
            return self._uid

    def user_uid(self):
        """Return the UID of the user who created this authorisation"""
        if self.is_null():
            return None
        else:
            return self._user_uid

    def user_guid(self):
        """Return the global UID for this user"""
        return "%s@%s" % (self.user_uid(), self.identity_uid())

    def identifiers(self):
        """Return a dictionary of the full set of identifiers attached
           to this authorisation (e.g. user_guid, group_guid(s) etc.)
        """
        return {"user_guid": self.user_guid()}

    def session_uid(self):
        """Return the login session that authenticated the user"""
        if self.is_null():
            return None
        else:
            return self._session_uid

    def identity_url(self):
        """Return the URL of the identity service that authenticated
           the user
        """
        if self.is_null():
            return None
        else:
            return self._identity_url

    def identity_uid(self):
        """Return the UID of the identity service that authenticated
           the user
        """
        if self.is_null():
            return None
        else:
            return self._identity_uid

    def signature_time(self):
        """Return the time when the authentication was signed"""
        if self.is_null():
            return None
        else:
            return self._auth_datetime

    def last_verification_time(self):
        """Return the last time this authorisation was verified. Note that
           you should re-verify authorisations periodically, to ensure that
           they identity service is still happy that the login session was
           not suspicious
        """
        if self.is_null():
            return None
        else:
            return self._last_validated_datetime

    def last_verified_resource(self):
        """Return the resource that was used for the last successful
           verification of this authorisation. This returns None
           if this has not been verified before
        """
        try:
            return self._last_verified_resource
        except:
            return None

    def signature(self):
        """Return the actual signature"""
        if self.is_null():
            return None
        else:
            return self._signature

    def is_stale(self, stale_time=7200):
        """Return whether or not this authorisation is stale.
           'stale_time' is the number of seconds after which
           the authorisation is considered stale (and thus
           no longer valid)
        """
        stale_time = self._fix_integer(stale_time, 365*24*7200)

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now

        now = _get_datetime_now()

        if now >= self._auth_datetime:
            return ((now - self._auth_datetime).seconds > stale_time)
        else:
            # datetime returns large positive numbers if time is
            # in the future - expect a little difference if client 
            # clock is fast. Give up to 10 seconds of leeway
            return (self._auth_datetime - now).seconds > 10

    def _get_user_public_cert(self, scope=None, permissions=None):
        """Internal function that returns the public certificate
           of the user who signed this authorisation. This will
           check that the authorisation was not signed after the
           user logged out, as well as validating the services
           that provide the user session keys etc.
        """
        must_fetch = False

        try:
            if scope != self._scope or permissions != self._permissions:
                must_fetch = True
        except:
            must_fetch = True

        if self._pubcert is not None:
            if not must_fetch:
                try:
                    return self._pubcert
                except:
                    pass

        try:
            testing_key = self._testing_key
        except:
            testing_key = None

        if testing_key is not None:
            if not self._is_testing:
                raise PermissionError(
                    "You cannot pass a test key to a non-testing "
                    "Authorisation")

            return testing_key

        # we need to get the public signing key for this session
        from Acquire.Service import get_trusted_service \
            as _get_trusted_service
        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now

        try:
            identity_service = _get_trusted_service(self._identity_url)
        except:
            raise PermissionError(
                "Unable to verify the authorisation as we do not trust "
                "the identity service at %s" % self._identity_url)

        if not identity_service.can_identify_users():
            raise PermissionError(
                "Cannot verify an Authorisation that does not use a "
                "valid identity service")

        if identity_service.uid() != self._identity_uid:
            raise PermissionError(
                "Cannot auth_once this Authorisation as the actual UID of "
                "the identity service at '%s' (%s) does not match "
                "the UID of the service that signed this authorisation "
                "(%s)" % (self._identity_url, identity_service.uid(),
                          self._identity_uid))

        response = identity_service.get_session_info(
                                session_uid=self._session_uid,
                                scope=scope, permissions=permissions)

        try:
            user_uid = response["user_uid"]
        except:
            pass

        if self._user_uid != user_uid:
            raise PermissionError(
                "Cannot verify the authorisation as there is "
                "disagreement over the UID of the user who signed "
                "the authorisation. %s versus %s" %
                (self._user_uid, user_uid))

        try:
            logout_datetime = _string_to_datetime(
                                    response["logout_datetime"])
        except:
            logout_datetime = None

        if logout_datetime:
            # the user has logged out from this session - ensure that
            # the authorisation was created before the user logged out
            if logout_datetime < self.signature_time():
                raise PermissionError(
                    "This authorisation was signed after the user logged "
                    "out. This means that the authorisation is not valid. "
                    "Please log in again and create a new authorisation.")

        from Acquire.Crypto import PublicKey as _PublicKey
        pubcert = _PublicKey.from_data(response["public_cert"])

        self._pubcert = pubcert
        self._scope = scope
        self._permissions = permissions
        return pubcert

    def assert_once(self, stale_time=7200, scope=None,
                    permissions=None):
        """Assert that this is in the one and only time that this
           service has seen this authorisation. This records the
           UID of the authorisation to the object store and then
           verifies that the signature of the UID is correct.

           There is a small race condition if the service asserts
           the authorisation at the exact same time, but this is
           a highly unlikely occurance. The aim is to prevent
           replay attacks.
        """
        if self.is_null():
            raise PermissionError("Cannot assert_once a null Authorisation")

        if self.is_stale(stale_time):
            if now < self._auth_datetime:
                raise PermissionError("Cannot assert_once an Authorisation signed "
                                      "in the future - please check your clock")
            else:
                raise PermissionError("Cannot assert_once a stale Authorisation")

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import get_datetime_now_to_string \
            as _get_datetime_now_to_string

        bucket = _get_service_account_bucket()
        authkey = "auth_once/%s" % self._uid
        now = _get_datetime_now_to_string()

        try:
            data = _ObjectStore.get_string_object(bucket=bucket, key=authkey)
        except:
            data = None

        if data is not None:
            raise PermissionError(
                "Cannot auth_once the authorisation as it has been used "
                "before on this service!")

        # This is the first time this authorisation has been seen.
        # Record this to the object store to prevent anyone else
        # from using this authorisation on this service. There is a
        # small race condition here, but this would be extremely
        # challenging to exploit, and mitigating it would be a
        # significant performance problem. Ideally the object store
        # would have a "test_and_set" to enable us to set only if
        # the previous value is None
        _ObjectStore.set_string_object(bucket=bucket, key=authkey,
                                       string_data=now)

        # Now validate that the signature of the UID is correct
        public_cert = self._get_user_public_cert(scope=scope,
                                                 permissions=permissions)

        if public_cert is None:
            raise PermissionError(
                "There is no public certificate for this user in "
                "scope '%s' with permissions '%s'" % (scope, permissions))

        try:
            public_cert.verify(self._siguid, self._uid)
        except Exception as e:
            raise PermissionError(
                "Cannot auth_once the authorisation as the signature "
                "is invalid! % s" % str(e))

    def is_verified(self, refresh_time=3600, stale_time=7200):
        """Return whether or not this authorisation has been verified. Note
           that this will cache any verification for 'refresh_time' (in
           seconds)

           'stale_time' gives the time (in seconds) beyond which the
           authorisation will be considered stale (and thus not valid).
           By default this is 7200 seconds (2 hours), meaning that the
           authorisation must be used within 2 hours to be valid.
        """
        refresh_time = self._fix_integer(refresh_time, 24*3600)

        from Acquire.ObjectStore import get_datetime_now \
            as _get_datetime_now

        now = _get_datetime_now()

        if self._last_validated_datetime is not None:
            if (now - self._last_validated_datetime).seconds < refresh_time:
                # no need to re-validate
                return not self.is_stale(stale_time)

        return False

    def verify(self, resource=None, refresh_time=3600, stale_time=7200,
               force=False, accept_partial_match=False,
               scope=None, permissions=None, return_identifiers=True):
        """Verify that this is a valid authorisation provided by the
           user for the passed 'resource'. This will
           cache the verification for 'refresh_time' (in seconds), but
           re-verification can be forced if 'force' is True.

           'stale_time' gives the time (in seconds) beyond which the
           authorisation will be considered stale (and thus not valid).
           By default this is 7200 seconds (2 hours), meaning that the
           authorisation must be used within 2 hours to be valid.

           If 'accept_partial_match' is True, then if this Authorisation
           has been previously validated, then this previous authorisation
           is valid if the previously-verified resource contains
           'resource', e.g. if you have previously verified that
           "create ABC123" is the verified resource, then this will
           still verify if "ABC123" if the partially-accepted match

           If 'scope' is passed, then verify that the user logged in
           and signed the authorisation with the required 'scope'.

           If 'permissions' is passed, then verify that the user
           logged in and signed the authorisation with at least
           the specified 'permissions'

           If 'testing_key' is passed, then this object is being
           tested as part of the unit tests

           If the authorisation was verified, then if 'return_identifiers'
           is True then this will return the full set of identifiers
           associated with the user who provided the authorisation
        """
        if self.is_null():
            raise PermissionError("Cannot verify a null Authorisation")

        if self.is_stale(stale_time):
            now = _get_datetime_now()
            if now < self._auth_datetime:
                raise PermissionError("Cannot verify an Authorisation signed "
                                      "in the future - please check your clock")
            else:
                raise PermissionError("Cannot verify a stale Authorisation")

        matched_resource = False

        try:
            last_resource = self._last_verified_resource
        except:
            last_resource = None

        if last_resource is not None:
            if accept_partial_match:
                if resource is None:
                    matched_resource = True
                else:
                    matched_resource = (last_resource.find(resource) != -1)
            else:
                matched_resource = (resource == last_resource)

        if not force:
            if self.is_verified(refresh_time=refresh_time,
                                stale_time=stale_time):
                if matched_resource:
                    if return_identifiers:
                        return self.identifiers()
                    else:
                        return

        # Now validate that the signature of the UID is correct
        public_cert = self._get_user_public_cert(scope=scope,
                                                 permissions=permissions)

        message = self._get_message(resource=resource,
                                    matched_resource=matched_resource)

        try:
            public_cert.verify(self._signature, message)
        except:
            raise PermissionError(
                "Cannot verify the authorisation as the signature "
                "for resource '%s' is invalid!" % resource)

        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now
        self._last_validated_datetime = _get_datetime_now()
        self._last_verified_resource = resource
        self._last_verified_key = public_cert

        if return_identifiers:
            return self.identifiers()
        else:
            return

    @staticmethod
    def from_data(data):
        """Return an authorisation created from the json-decoded dictionary"""
        auth = Authorisation()

        if (data and len(data) > 0):
            from Acquire.ObjectStore import string_to_datetime \
                as _string_to_datetime
            from Acquire.ObjectStore import string_to_bytes \
                as _string_to_bytes

            auth._user_uid = data["user_uid"]
            auth._session_uid = data["session_uid"]
            auth._identity_url = data["identity_url"]
            auth._identity_uid = data["identity_uid"]
            auth._uid = data["uid"]
            parts = auth._uid.split("/")
            auth._auth_datetime = _string_to_datetime(parts[0])
            auth._signature = _string_to_bytes(data["signature"])
            auth._siguid = _string_to_bytes(data["siguid"])
            auth._last_validated_datetime = None

            if "is_testing" in data:
                auth._is_testing = data["is_testing"]

        return auth

    def to_data(self):
        """Return this object serialised to a json-encoded dictionary"""
        data = {}

        if self.is_null():
            return data

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import bytes_to_string \
            as _bytes_to_string

        data["user_uid"] = str(self._user_uid)
        data["session_uid"] = str(self._session_uid)
        data["identity_url"] = str(self._identity_url)
        data["identity_uid"] = str(self._identity_uid)
        data["uid"] = self._uid
        data["signature"] = _bytes_to_string(self._signature)
        data["siguid"] = _bytes_to_string(self._siguid)

        try:
            data["is_testing"] = self._is_testing
        except:
            pass

        return data




import datetime as _datetime
import json as _json
import os as _os

__all__ = ["OSPar", "BucketReader", "BucketWriter",
           "ObjectReader", "ObjectWriter"]


class OSPar:
    """This class holds the result of a pre-authenticated request
       to access a bucket or object in an object store
       (a Object Store PAR - also called a pre-signed request). This holds a
       pre-authenticated URL to access either;

       (1) A individual object in an object store (read or write)
       (2) An entire bucket in the object store (write only)
       (3) A calculation to be performed on the compute service (start or stop)

       The OSPar is created encrypted, so can only be used by the
       person or service that has access to the decryption key

       This is a low-level object that should be used with care
       when manipulating data in the object store. This class
       gives direct access to objects or buckets in the object
       store to remote clients. It is designed to enable the
       higher-level Drive and File classes to directly read and
       write data without having to go through a serverless
       function or having to gain login-access to the object store

       Args:
            url (str, default=None): URL for access
            key (str, default=None): Key for request
            encrypt_key (str, default=None): Key used to encrypt
            the OSPar
            expires_datatime (datetime, default=None): Datetime at
            which the OSPar expires
            is_readable (bool, default=False): If True read access granted
            is_writeable (bool, default=False): If True write access granted
            is_executable (bool, default=False): If True OSPar triggers a
            calculation
            driver_details (str, default=None): Contains extra details for
            OSPar creation

    """
    def __init__(self, url=None, key=None,
                 encrypt_key=None,
                 expires_datetime=None,
                 is_readable=False,
                 is_writeable=False,
                 driver_details=None):
        """Construct an OSPar result by passing in the URL at which the
           object can be accessed, the UTC datetime when this expires,
           whether this is readable or writeable, and
           the encryption key to use to encrypt the OSPar.

           You can optionally pass in the key for the object in the
           object store that
           this provides access to. If this is not supplied, then an
           entire bucket is accessed). If 'is_readable', then read-access
           has been granted, while if 'is_writeable' then write
           access has been granted.

           Otherwise no access is possible.

           driver_details is provided by the machinery that creates
           the OSPar, and supplies extra details that are used by the
           driver to create, register and manage OSPars... You should
           not do anything with driver_details yourself
        """
        service_url = None

        if url is None:
            is_readable = True
            self._uid = None
        else:
            from Acquire.Crypto import PublicKey as _PublicKey
            from Acquire.Crypto import PrivateKey as _PrivateKey

            if isinstance(encrypt_key, _PrivateKey):
                encrypt_key = encrypt_key.public_key()

            if not isinstance(encrypt_key, _PublicKey):
                raise TypeError(
                    "You must supply a valid PublicKey to encrypt a OSPar")

            url = encrypt_key.encrypt(url)

            from Acquire.ObjectStore import create_uid as _create_uid
            self._uid = _create_uid()

            try:
                from Acquire.Service import get_this_service \
                    as _get_this_service
                service_url = _get_this_service().canonical_url()
            except:
                pass

        self._url = url
        self._key = key
        self._expires_datetime = expires_datetime
        self._service_url = service_url

        if driver_details is not None:
            if not isinstance(driver_details, dict):
                raise TypeError("The driver details must be a dictionary")

        self._driver_details = driver_details

        if is_readable:
            self._is_readable = True
        else:
            self._is_readable = False

        if is_writeable:
            self._is_writeable = True
        else:
            self._is_writeable = False

        if not (self._is_readable or self._is_writeable):
            from Acquire.Client import PARPermissionsError
            raise PARPermissionsError(
                "You cannot create a OSPar that has no read or write "
                "permissions!")

    def __str__(self):
        if self.seconds_remaining() < 1:
            return "OSPar( expired )"

        if self._key is None:
            return "OSPar( bucket=True, seconds_remaining=%s )" % \
                (self.seconds_remaining(buffer=0))
        else:
            return "OSPar( key=%s, seconds_remaining=%s )" % \
                (self.key(), self.seconds_remaining(buffer=0))

    def _set_private_key(self, privkey):
        """Call this function to set the private key for this
           OSPar. This is the private key that is used to
           decrypt the OSPar, and is provided here if you want
           to use the OSPar without having to always supply
           the key (by definition, you are the only person
           who has the key)

           Args:
                privkey (str): Private key to use for this OSPar
           Returns:
                None
        """
        from Acquire.Crypto import PrivateKey as _PrivateKey

        if not isinstance(privkey, _PrivateKey):
            raise TypeError("The private key must be type PrivateKey")

        self._privkey = privkey

    def _get_privkey(self, decrypt_key=None):
        """Return the private key used to decrypt the OSPar, passing in
           the user-supplied key if needed

           Args:
                decrypt_key (str): Key to decrypt the OSPar
           Returns:
                str: Key for decryption of OSPar
        """
        try:
            if self._privkey is not None:
                return self._privkey
        except:
            pass

        if decrypt_key is None:
            raise PermissionError(
                "You must supply a private key to decrypt this OSPar")

        from Acquire.Crypto import PrivateKey as _PrivateKey
        if not isinstance(decrypt_key, _PrivateKey):
            raise TypeError("The supplied private key must be type PrivateKey")

        return decrypt_key

    def is_null(self):
        """Return whether or not this is null

           Returns:
                bool: True if OSPar is null, else False
        """
        return self._uid is None

    @staticmethod
    def checksum(data):
        """Return the checksum of the passed data. This is used either
           for validating data, and is also used to create a checksum
           of the URL so that the user can demonstrate that they can
           decrypt this OSPar

           Args:
                data (str): Data to checksum
           Returns:
                str: MD5 checksum of data
        """
        from hashlib import md5 as _md5
        md5 = _md5()

        if isinstance(data, str):
            data = data.encode("utf-8")

        md5.update(data)
        return md5.hexdigest()

    def url(self, decrypt_key=None):
        """Return the URL at which the bucket/object can be accessed. This
           will raise a PARTimeoutError if the url has less than 30 seconds
           of validity left. Note that you must pass in the key used to
           decrypt the OSPar

           Args:
                decrypt_key (str, default=None): Key for decryption of data
           Returns:
                TODO - ensure this is correct
                str: Decrypted data

           """
        if self.seconds_remaining(buffer=30) <= 0:
            from Acquire.Client import PARTimeoutError
            raise PARTimeoutError(
                "The URL behind this OSPar has expired and is no longer valid")

        return self._get_privkey(decrypt_key).decrypt(self._url)

    def service_url(self):
        """Return the URL of the service that created this OSPar

           Returns:
                str: URL of service that created this OSPar
        """
        if self.is_null():
            return None
        else:
            return self._service_url

    def service(self):
        """Return the service that created this OSPar

           Returns:
                Service: Service object that created this OSPar
        """
        service_url = self.service_url()

        if service_url is not None:
            from Acquire.Service import get_trusted_service \
                as _get_trusted_service

            print(service_url)
            return _get_trusted_service(service_url=service_url)
        else:
            return None

    def uid(self):
        """Return the UID of this OSPar

           Returns:
                str: UID of this OSPar
        """
        return self._uid

    def driver_details(self):
        """Return the driver details for this OSPar. This is used only
           on the service that created the OSPar, and returns an empty
           dictionary if the details are not available

           Returns:
                str: Driver details for this OSPar
        """
        if self._driver_details is None:
            return {}
        else:
            return self._driver_details

    def driver(self):
        """Return the driver behind this OSPar - this is only available on
           the service

           Returns:
                str: Driver behind this OSPar
        """
        try:
            return self.driver_details()["driver"]
        except:
            return None

    def fingerprint(self):
        """Return a fingerprint for this OSPar that can be used
           in authorisations

           Returns:
                str: UID for this OSPar
        """
        return self._uid

    def is_readable(self):
        """Return whether or not this OSPar gives read access

           Returns:
                bool: True if OSPar gives read access
        """
        return self._is_readable

    def is_writeable(self):
        """Return whether or not this OSPar gives write access

           Returns:
                bool: True if OSPar gives write access

        """
        return self._is_writeable

    def key(self):
        """Return the key for the object this accesses - this is None
           if the OSPar grants access to the entire bucket

           Returns:
                str: Key for the object this accesses

           """
        return self._key

    def is_bucket(self):
        """Return whether or not this OSPar is for an entire bucket

           Returns:
                bool: True if this OSPar is for a bucket, else False

        """
        return self._key is None

    def is_object(self):
        """Return whether or not this OSPar is for a single object

           Returns:
                bool: True if OSPar is for a single object, else False
        """
        return self._key is not None

    def expires_when(self):
        """Return when this OSPar expires (or expired)

           Returns:
                datetime: Time at which this OSPar expires

        """
        if self.is_null():
            return None
        else:
            return self._expires_datetime

    def seconds_remaining(self, buffer=30):
        """Return the number of seconds remaining before this OSPar expires.
           This will return 0 if the OSPar has already expired. To be safe,
           you should renew PARs if the number of seconds remaining is less
           than 60. This will subtract 'buffer' seconds from the actual
           validity to provide a buffer against race conditions (function
           says this is valid when it is not)

           Args:
                buffer (int, default=30): buffer OSPar validity (seconds)
           Returns:
                datetime: Seconds remaining on OSPar validity
        """
        if self.is_null():
            return 0

        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now

        buffer = float(buffer)

        if buffer < 0:
            buffer = 0

        now = _get_datetime_now()

        delta = (self._expires_datetime - now).total_seconds() - buffer

        if delta < 0:
            return 0
        else:
            return delta

    def read(self, decrypt_key=None):
        """Return an object that can be used to read data from this OSPar

           Args:
                decrypt_key (str, default=None): Key to decrypt this OSPar
           Returns:
                BucketReader or ObjectReader: Object that can be used to
                read this OSPar
        """
        if not self.is_readable():
            from Acquire.Client import PARPermissionsError
            raise PARPermissionsError(
                "You do not have permission to read from this OSPar: %s"
                % self)

        if self.is_bucket():
            return BucketReader(self, self._get_privkey(decrypt_key))
        else:
            return ObjectReader(self, self._get_privkey(decrypt_key))

    def write(self, decrypt_key=None):
        """Return an object that can be used to write data to this OSPar

           Args:
                decrypt_key (str, default=None): Key to decrypt this OSPar
           Returns:
                BucketReader or ObjectReader: Object that can be used to
                write to this OSPar

        """
        if not self.is_writeable():
            from Acquire.Client import PARPermissionsError
            raise PARPermissionsError(
                "You do not have permission to write to this OSPar: %s" % self)

        if self.is_bucket():
            return BucketWriter(self, self._get_privkey(decrypt_key))
        else:
            return ObjectWriter(self, self._get_privkey(decrypt_key))

    def close(self, decrypt_key=None):
        """Close this OSPar - this closes and deletes the OSPar. You must
           pass in the decryption key so that you can validate that
           you have permission to read (and thus close) this OSPar

           Args:
                decrypt_key (str, default=None): Key to decrypt this OSPar
           Returns:
                None
        """
        if self.is_null():
            return

        service = self.service()

        if service is not None:
            # we confirm we have permission to close this OSPar by sending
            # a checksum of the url (which only we would know)
            url = self.url(decrypt_key=decrypt_key)

            args = {"par_uid": self._uid,
                    "url_checksum": OSPar.checksum(url)}

            service.call_function(function="close_ospar", args=args)

        # now that the OSPar is closed, set it into a null state
        import copy as _copy
        par = OSPar()
        self.__dict__ = _copy.copy(par.__dict__)

    def to_data(self, passphrase=None):
        """Return a json-serialisable dictionary that contains all data
           for this object

           Args:
                passphrase (str, default=None): Passphrase to use to
                encrypt OSPar
           Returns:
                dict: JSON serialisable dictionary
        """
        data = {}

        if self._url is None:
            return data

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import bytes_to_string \
            as _bytes_to_string

        data["url"] = _bytes_to_string(self._url)
        data["uid"] = self._uid
        data["key"] = self._key
        data["expires_datetime"] = _datetime_to_string(self._expires_datetime)
        data["is_readable"] = self._is_readable
        data["is_writeable"] = self._is_writeable

        try:
            if self._service_url is not None:
                data["service_url"] = self._service_url
        except:
            pass

        try:
            privkey = self._privkey
        except:
            privkey = None

        if privkey is not None:
            if passphrase is not None:
                data["privkey"] = privkey.to_data(passphrase)

        # note that we don't save the driver details as these
        # are stored separately on the service

        return data

    @staticmethod
    def from_data(data, passphrase=None):
        """Return a OSPar constructed from the passed json-deserliased
           dictionary

           Args:
                data (dict): JSON-deserialised dictionary from which to
                create OSPar
            Returns:
                OSPar: OSPar object created from dict
        """
        if data is None or len(data) == 0:
            return OSPar()

        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime
        from Acquire.ObjectStore import string_to_bytes \
            as _string_to_bytes

        par = OSPar()

        par._url = _string_to_bytes(data["url"])
        par._key = data["key"]
        par._uid = data["uid"]

        if par._key is not None:
            par._key = str(par._key)

        par._expires_datetime = _string_to_datetime(data["expires_datetime"])
        par._is_readable = data["is_readable"]
        par._is_writeable = data["is_writeable"]

        if "service_url" in data:
            par._service_url = data["service_url"]

        if "privkey" in data:
            if passphrase is not None:
                from Acquire.Crypto import PrivateKey as _PrivateKey
                par._privkey = _PrivateKey.from_data(data["privkey"],
                                                     passphrase)

        # note that we don't load the driver details as this
        # is stored and loaded separately on the service

        return par


def _url_to_filepath(url):
    """Internal function used to strip the "file://" from the beginning
       of a file url

       Args:
            url (str): URL to clean
       Returns:
            str: URL string with file:// removed
    """
    return url[7:]


def _read_local(url):
    """Internal function used to read data from the local testing object
       store

       Args:
            url (str): URL from which to read data
       Returns:
            bytes: Data read from file
    """
    with open("%s._data" % _url_to_filepath(url), "rb") as FILE:
        return FILE.read()


def _read_remote(url):
    """Internal function used to read data from a remote URL

       Args:
            url (str): Remote URL from which to read data
       Returns:
            str: HTTP request content

    """
    status_code = None
    response = None

    try:
        from Acquire.Stubs import requests as _requests
        response = _requests.get(url)
        status_code = response.status_code
    except Exception as e:
        from Acquire.Client import PARReadError
        raise PARReadError(
            "Cannot read the remote OSPar URL '%s' because of a possible "
            "nework issue: %s" % (url, str(e)))

    output = response.content

    if status_code != 200:
        from Acquire.Client import PARReadError
        raise PARReadError(
            "Failed to read data from the OSPar URL. HTTP status code = %s, "
            "returned output: %s" % (status_code, output))

    return output


def _list_local(url):
    """Internal function to list all of the objects keys below 'url'

       Args:
            url (str): URL from which to read data
       Returns:
            list: List of object keys
    """
    local_dir = _url_to_filepath(url)

    keys = []

    for dirpath, _, filenames in _os.walk(local_dir):
        local_path = dirpath[len(local_dir):]
        has_local_path = (len(local_path) > 0)

        for filename in filenames:
            if filename.endswith("._data"):
                filename = filename[0:-6]

                if has_local_path:
                    keys.append("%s/%s" % (local_path, filename))
                else:
                    keys.append(filename)

    return keys


def _list_remote(url):
    """Internal function to list all of the objects keys below 'url'

       Currently unimplemented

       Args:
            url (str): URL from which to read data
       Returns:
            list: Empty list
    """
    return []


def _write_local(url, data):
    """Internal function used to write data to a local file

       Args:
            url (str): URL to write data to
            data (bytes): Data to write
       Returns:
            None
    """
    filename = "%s._data" % _url_to_filepath(url)

    try:
        with open(filename, 'wb') as FILE:
            FILE.write(data)
            FILE.flush()
    except:
        dir = "/".join(filename.split("/")[0:-1])
        _os.makedirs(dir, exist_ok=True)
        with open(filename, 'wb') as FILE:
            FILE.write(data)
            FILE.flush()


def _write_remote(url, data):
    """Internal function used to write data to the passed remote URL

       Args:
            url (str): Remote URL to write data to
            data (bytes): Data to write
       Returns:
            None
    """
    try:
        from Acquire.Stubs import requests as _requests
        response = _requests.put(url, data=data)
        status_code = response.status_code
    except Exception as e:
        from Acquire.Client import PARWriteError
        raise PARWriteError(
            "Cannot write data to the remote OSPar URL '%s' because of a "
            "possible nework issue: %s" % (url, str(e)))

    if status_code != 200:
        from Acquire.Client import PARWriteError
        raise PARWriteError(
            "Cannot write data to the remote OSPar URL '%s' because of a "
            "possible nework issue: %s" % (url, str(response.content)))


def _join_bucket_and_prefix(url, prefix):
    """Join together the passed url and prefix, returning the
       url directory and the remainig part which is the start
       of the file name

       Args:
            url (str): URL to combine
            prefix (str): Prefix to combine
       Returns:
            str: URL and prefix processed concatenated
    """
    if prefix is None:
        return url

    parts = prefix.split("/")

    return ("%s/%s" % (url, "/".join(parts[0:-2])), parts[-1])


class BucketReader:
    """This class provides functions to enable reading data from a
       bucket via a OSPar

       Args:
            par (OSPar, default=None): OSPar to use for data access
            decrypt_key (str, default=None): Key to decrypt data in bucket

    """
    def __init__(self, par=None, decrypt_key=None):
        if par:
            if not isinstance(par, OSPar):
                raise TypeError(
                    "You can only create a BucketReader from a OSPar")
            elif not par.is_bucket():
                raise ValueError(
                    "You can only create a BucketReader from a OSPar that "
                    "represents an entire bucket: %s" % par)
            elif not par.is_readable():
                from Acquire.Client import PARPermissionsError
                raise PARPermissionsError(
                    "You cannot create a BucketReader from a OSPar without "
                    "read permissions: %s" % par)

            self._par = par
            self._url = par.url(decrypt_key)
        else:
            self._par = None

    def get_object(self, key):
        """Return the binary data contained in the key 'key' in the
           passed bucket

           Args:
                key (str): Key to access data in bucket
           Returns:
                bytes: Data referred to by key
        """
        if self._par is None:
            from Acquire.Client import PARError
            raise PARError("You cannot read data from an empty OSPar")

        while key.startswith("/"):
            key = key[1:]

        url = self._url

        if url.endswith("/"):
            url = "%s%s" % (url, key)
        else:
            url = "%s/%s" % (url, key)

        if url.startswith("file://"):
            return _read_local(url)
        else:
            return _read_remote(url)

    def get_object_as_file(self, key, filename):
        """Get the object contained in the key 'key' in the passed 'bucket'
           and writing this to the file called 'filename'

           Args:
                key (str): Key to access data in bucket
           Returns:
                None
        """
        objdata = self.get_object(key)

        with open(filename, "wb") as FILE:
            FILE.write(objdata)

    def get_string_object(self, key):
        """Return the string in 'bucket' associated with 'key'

           Args:
                key (str): Key to access data in bucket
           Returns:
                str: String referred to by key
        """
        data = self.get_object(key)

        try:
            return data.decode("utf-8")
        except Exception as e:
            raise TypeError(
                "The object behind this OSPar cannot be converted to a "
                "string. Error is: %s" % str(e))

    def get_object_from_json(self, key):
        """Return an object constructed from json stored at 'key' in
           the passed bucket. This raises an exception if there is no
           data or the OSPar has expired
        """
        data = self.get_string_object(key)
        return _json.loads(data)

    def get_all_object_names(self, prefix=None):
        """Returns the names of all objects in the passed bucket"""
        (url, part) = _join_bucket_and_prefix(self._url, prefix)

        if url.startswith("file://"):
            objnames = _list_local(url)
        else:
            objnames = _list_remote(url)

        # scan the object names returned and discard the ones that don't
        # match the prefix
        matches = []

        if len(part) > 0:
            for objname in objnames:
                if objname.startswith(part):
                    objname = objname[len(part):]

                    while objname.startswith("/"):
                        objname = objname[1:]

                    matches.append(objname)
        else:
            matches = objnames

        return matches

    def get_all_objects(self, prefix=None):
        """Return all of the objects in the passed bucket"""
        names = self.get_all_object_names(prefix)

        objects = {}

        if prefix:
            for name in names:
                objects[name] = self.get_object(
                                    "%s/%s" % (prefix, name))
        else:
            for name in names:
                objects[name] = self.get_object(name)

        return objects

    def get_all_strings(self, prefix=None):
        """Return all of the strings in the passed bucket"""
        objects = self.get_all_objects(prefix)

        names = list(objects.keys())

        for name in names:
            try:
                s = objects[name].decode("utf-8")
                objects[name] = s
            except:
                del objects[name]

        return objects


class BucketWriter:
    """This class provides functions to enable writing data to a
       bucket via a OSPar
    """
    def __init__(self, par=None, decrypt_key=None):
        if par:
            if not isinstance(par, OSPar):
                raise TypeError(
                    "You can only create a BucketReader from a OSPar")
            elif not par.is_bucket():
                raise ValueError(
                    "You can only create a BucketReader from a OSPar that "
                    "represents an entire bucket: %s" % par)
            elif not par.is_writeable():
                from Acquire.Client import PARPermissionsError
                raise PARPermissionsError(
                    "You cannot create a BucketWriter from a OSPar without "
                    "write permissions: %s" % par)

            self._par = par
            self._url = par.url(decrypt_key)
        else:
            self._par = None

    def set_object(self, key, data):
        """Set the value of 'key' in 'bucket' to binary 'data'"""
        if self._par is None:
            from Acquire.Client import PARError
            raise PARError("You cannot write data to an empty OSPar")

        while key.startswith("/"):
            key = key[1:]

        url = self._url

        if url.endswith("/"):
            url = "%s%s" % (url, key)
        else:
            url = "%s/%s" % (url, key)

        if url.startswith("file://"):
            return _write_local(url, data)
        else:
            return _write_remote(url, data)

    def set_object_from_file(self, key, filename):
        """Set the value of 'key' in 'bucket' to equal the contents
           of the file located by 'filename'"""
        with open(filename, "rb") as FILE:
            data = FILE.read()
            self.set_object(key, data)

    def set_string_object(self, key, string_data):
        """Set the value of 'key' in 'bucket' to the string 'string_data'"""
        self.set_object(key, string_data.encode("utf-8"))

    def set_object_from_json(self, key, data):
        """Set the value of 'key' in 'bucket' to equal to contents
           of 'data', which has been encoded to json"""
        self.set_string_object(key, _json.dumps(data))


class ObjectReader:
    """This class provides functions for reading an object via a OSPar"""
    def __init__(self, par=None, decrypt_key=None):
        if par:
            if not isinstance(par, OSPar):
                raise TypeError(
                    "You can only create an ObjectReader from a OSPar")
            elif par.is_bucket():
                raise ValueError(
                    "You can only create an ObjectReader from a OSPar that "
                    "represents an object: %s" % par)
            elif not par.is_readable():
                from Acquire.Client import PARPermissionsError
                raise PARPermissionsError(
                    "You cannot create an ObjectReader from a OSPar without "
                    "read permissions: %s" % par)

            self._par = par
            self._url = par.url(decrypt_key)
        else:
            self._par = None

    def get_object(self):
        """Return the binary data contained in this object"""
        if self._par is None:
            from Acquire.Client import PARError
            raise PARError("You cannot read data from an empty OSPar")

        url = self._url

        if url.startswith("file://"):
            return _read_local(url)
        else:
            return _read_remote(url)

    def get_object_as_file(self, filename):
        """Get the object contained in this OSPar and write this to
           the file called 'filename'"""
        objdata = self.get_object()

        with open(filename, "wb") as FILE:
            FILE.write(objdata)

    def get_string_object(self):
        """Return the object behind this OSPar as a string (raises exception
           if it is not a string)'"""
        data = self.get_object()

        try:
            return data.decode("utf-8")
        except Exception as e:
            raise TypeError(
                "The object behind this OSPar cannot be converted to a "
                "string. Error is: %s" % str(e))

    def get_object_from_json(self):
        """Return an object constructed from json stored at behind
           this OSPar. This raises an exception if there is no data
           or the OSPar has expired
        """
        data = self.get_string_object()
        return _json.loads(data)


class ObjectWriter(ObjectReader):
    """This is an extension of ObjectReader that also allows writing to
       the object via the OSPar
    """
    def __init__(self, par=None, decrypt_key=None):
        if par:
            if not isinstance(par, OSPar):
                raise TypeError(
                    "You can only create an ObjectReader from a OSPar")
            elif par.is_bucket():
                raise ValueError(
                    "You can only create an ObjectReader from a OSPar that "
                    "represents an object: %s" % par)
            elif not par.is_writeable():
                from Acquire.Client import PARPermissionsError
                raise PARPermissionsError(
                    "You cannot create an ObjectWriter from a OSPar without "
                    "write permissions: %s" % par)

            self._par = par
            self._url = par.url(decrypt_key)
        else:
            self._par = None

    def set_object(self, data):
        """Set the value of the object behind this OSPar to the binary
           'data'
        """
        if self._par is None:
            from Acquire.Client import PARError
            raise PARError("You cannot write data to an empty OSPar")

        url = self._url

        if url.startswith("file://"):
            return _write_local(url, data)
        else:
            return _write_remote(url, data)

    def set_object_from_file(self, filename):
        """Set the value of the object behind this OSPar to equal the contents
           of the file located by 'filename'"""
        with open(filename, "rb") as FILE:
            data = FILE.read()
            self.set_object(data)

    def set_string_object(self, string_data):
        """Set the value of the object behind this OSPar to the
           string 'string_data'
        """
        self.set_object(string_data.encode("utf-8"))

    def set_object_from_json(self, data):
        """Set the value of the object behind this OSPar to equal to contents
           of 'data', which has been encoded to json"""
        self.set_string_object(_json.dumps(data))
